#All arguments are made keyword arguments for easier partial usage
def excess_lookup( pubkey=None, _hash=None, rtx=None, tx=None, excesses_storage=None):
  assert pubkey
  assert _hash
  assert rtx
  assert tx
  assert excesses_storage
  pk = pubkey.to_pubkey()
  _index = pk.serialize()+_hash
  return excess_lookup_by_index(_index, tx=tx, rtx=rtx, excesses_storage=excesses_storage)

def excess_lookup_by_index(index, tx=None, rtx=None, excesses_storage=None):
  excesses = tx.additional_excesses + list(tx.updated_excesses.values())
  excesses_indexes = {e.index:e for e in excesses}
  if index in excesses_indexes:
    return excesses_indexes[index].message
  e = excesses_storage.get_by_index(index, rtx=rtx)
  if e:
    return e.message
  return False




from time import time, sleep
from leer.wallet.key_manager import KeyManagerClass

from leer.core.lubbadubdub.ioput import IOput
from leer.core.lubbadubdub.address import Address
from leer.core.lubbadubdub.transaction import Transaction
from secp256k1_zkp import PrivateKey

from uuid import uuid4
import logging
import base64
     


logger = logging.getLogger("Wallet")

def wallet(syncer, config):
  '''
    Wallet is synchronous service which holds private keys and information about 
    owned outputs. It provides information for transactions and block templates
    generation.
  '''
  def get_height(timeout=2.5):
    _id = str(uuid4())
    syncer.queues['Notifications'].put({'action':'get', 'id':_id, 'key': 'blockchain height','sender': "Wallet"})
    message_queue = syncer.queues['Wallet']
    start_time = time()
    result = None
    while True:
      put_back = [] #We wait for specific message, all others will wait for being processed
      while not message_queue.empty():
        message = message_queue.get()
        if (not 'id' in message)  or (not message['id']==_id):
          put_back.append(message)
          continue
        result = message['result']
        break
      for message in put_back:
        message_queue.put(message)
      if result:
        break
      sleep(0.01)
      if time()-start_time>timeout:
        raise KeyError
    if result=='error':
      raise KeyError
    return result['value']

  notification_cache = {}
  def notify(key, value, timestamp=None):
    if (key in notification_cache) and (notification_cache[key]['value'] == value) and (time()-notification_cache[key]['timestamp'])<5:
      return #Do not spam notifications with the same values
    message = {}
    message['id'] = uuid4()
    message['sender'] = "Wallet"
    if not timestamp:
      timestamp = time()
    message['time'] = timestamp
    message['action']="set"
    message['key']=key
    message['value']=value
    syncer.queues["Notifications"].put(message)
    notification_cache[key] = {'value':value, 'timestamp':timestamp}


  #set logging
  default_log_level = logging.INFO;
  if "logging" in config:#debug, info, warning, error, critical
    loglevels = { "debug":logging.DEBUG, "info":logging.INFO, "warning":logging.WARNING, "error":logging.ERROR, "critical":logging.CRITICAL}
    if "base" in config["logging"] and config["logging"]["base"] in loglevels:
      logger.setLevel(loglevels[config["logging"]["base"]])
    if "wallet" in config["logging"] and config["logging"]["wallet"] in loglevels:
      #its ok to rewrite
      logger.setLevel(loglevels[config["logging"]["wallet"]])

  message_queue = syncer.queues['Wallet']
  _path = config['location']['wallet']
  km = KeyManagerClass(path=_path)
  notify('last wallet update', time())
  while True:
    sleep(0.01)
    while not message_queue.empty():
      message = message_queue.get()
      if 'action' in message:
        logger.info("Process message `%s`"% message['action'])
        logger.debug("Process message %s"%message)
      else:
        logger.info("Process message %s"%message)
      if not 'action' in message:
        continue
      if message['action']=="process new block":
        tx = Transaction(txos_storage=None, excesses_storage=None)
        tx.deserialize(message['tx'], rtx=None, skip_verification=True) #skip_verification allows us to not provide rtx
        block_height = message['height']
        last_time_updated = None
        with km.disc_txn(write = True) as w_txn:
          for index in tx.inputs:
            #Note it is not check whether output is unspent or not, we check that output is marked as our and unspent in our wallet
            if km.is_unspent(index, r_txn=w_txn):
              km.spend_output(index, block_height, w_txn=w_txn)
              last_time_updated = time()
          for _o in tx.outputs:
            if km.is_owned_pubkey(_o.address.pubkey.serialize(), r_txn=w_txn):
              km.add_output(_o, block_height, w_txn=w_txn)
              last_time_updated = time()
            if km.is_saved(_o, r_txn=w_txn):
              km.register_processed_output(_o.serialized_index, block_height, w_txn=w_txn)
        if last_time_updated:
          notify('last wallet update', last_time_updated)
      if message['action']=="process rollback":
        rollback = message['rollback_object']
        block_height = message['block_height']
        with km.disc_txn(write = True) as w_txn:
          km.rollback(block_height, w_txn=w_txn)
        last_time_updated = time()
        notify('last wallet update', last_time_updated)
      if message['action']=="process indexed outputs": #during private key import correspondent outputs will be processed again
        pass
      if message['action']=="give new taddress":
        with km.disc_txn(write = True) as w_txn:
          address = km.new_address(w_txn)
        response = {"id": message["id"], "result": address.to_text()}
        syncer.queues[message['sender']].put(response)
      if message['action']=="give new address":
        with km.disc_txn(write = True) as w_txn:
          address = km.new_address(w_txn)
        response = {"id": message["id"], "result": address.serialize()}
        syncer.queues[message['sender']].put(response)
      if message['action']=="get confirmed balance stats":
        response = {"id": message["id"]}
        try:
          height = get_height()
          with km.disc_txn(write = False) as r_txn:
            stats = km.get_confirmed_balance_stats(height, r_txn)
          response["result"] = stats
        except KeyError:
          response["result"] = "error: core_loop didn't set height yet"
        except Exception as e:
          response["result"] = "error: " +str(e)
        syncer.queues[message['sender']].put(response)
      if message['action']=="get confirmed balance list":
        response = {"id": message["id"]}
        try:
          height = get_height()
          with km.disc_txn(write = True) as r_txn:
            stats = km.get_confirmed_balance_list(height, r_txn)
          response["result"] = stats
        except KeyError:
          response["result"] = "error: core_loop didn't set height yet"
        except Exception as e:
          response["result"] = "error: " +str(e)
        syncer.queues[message['sender']].put(response)
      if message['action']=="give private key":
        pass
      if message['action']=="take private key":
        pass
      if message['action']=="give last transactions info":
        response = {"id": message["id"]}
        num = int(message["num"])
        with km.disc_txn(write = False) as r_txn:
          response["result"] = km.give_transactions(num, r_txn)
        syncer.queues[message['sender']].put(response) 
        continue
      if message['action']=="generate tx":
        response = {"id": message["id"]}
        value  = int(message["value"])
        taddress = message["address"]
        a = Address()
        a.from_text(taddress)
        try:
          current_height = get_height()
        except KeyError:
           response["result"] = "error"
           response["error"] = "core_loop didn't set height yet"
           syncer.queues[message['sender']].put(response)
           continue
        except Exception as e:
           response["result"] = "error"
           response["error"] = str(e)
           syncer.queues[message['sender']].put(response)
           continue
        with km.disc_txn(write = True) as w_txn:
          _list = km.get_confirmed_balance_list(current_height, r_txn=w_txn)
          list_to_spend = []
          summ = 0 
          utxos = []
          for address in _list:
              for texted_index in _list[address]:
                if summ>value+100000000: #TODO fee here
                  continue
                if isinstance(_list[address][texted_index], int):
                  _index = base64.b64decode(texted_index.encode())
                  ser_priv, ser_blinding, apc = km.get_output_private_data(_index, r_txn = w_txn)
                  priv = PrivateKey(ser_priv, raw=True)
                  blinding = PrivateKey(ser_blinding, raw=True)
                  utxos.append( (_index, _list[address][texted_index], priv, blinding, apc) )
                  summ+=_list[address][texted_index]
          if summ < value:
              response["result"] = "error"
              response["error"] = "Not enough matured coins"
              syncer.queues[message['sender']].put(response)
              continue
          tx = Transaction(None, None)
          tx.add_destination((a, value, True))
          tx.blindly_generate(km.new_address(w_txn), utxos, config["fee_policy"].get("generate_fee_per_kb", 3000))        
          km.save_generated_transaction(tx, w_txn)
          response["result"]=tx.serialize()
          syncer.queues[message['sender']].put(response)
      if message['action']=="stop":
        return



from leer.core.storage.txos_storage import TXOsStorage
from leer.core.storage.excesses_storage import ExcessesStorage
from leer.core.storage.headers_storage import HeadersStorage
from leer.core.chains.headers_manager import HeadersManager
from leer.core.storage.blocks_storage import BlocksStorage
from leer.core.chains.blockchain import Blockchain


from leer.core.lubbadubdub.transaction import Transaction
from leer.core.primitives.block import Block
from leer.core.primitives.block import ContextBlock
from leer.core.primitives.block import generate_genesis
from leer.core.storage.mempool_tx import MempoolTx
from leer.core.storage.utxo_index_storage import UTXOIndex

from leer.core.storage.storage_space import StorageSpace
import shutil, time, os

home = os.path.expanduser("~")
base_path = os.path.join(home, ".testleer")


def wipe_test_dirs():
  shutil.rmtree(base_path, True)


def rebuild_test_storage_space(path=base_path):
  global test_storage_space
  if not os.path.exists(path): 
      os.makedirs(path) #TODO catch

  try:
    if test_storage_space:
      del test_storage_space      
  except NameError:
    pass #it's ok: test_storage_space wasn't defined yet

  wipe_test_dirs()
  test_storage_space = StorageSpace(path)
  with test_storage_space.env.begin(write=True) as wtx:
    env = test_storage_space.env
    hs = HeadersStorage(test_storage_space, wtx)
    hm = HeadersManager(test_storage_space, do_not_check_pow=True)
    bs = BlocksStorage(test_storage_space, wtx)
    es = ExcessesStorage(test_storage_space, wtx)
    ts = TXOsStorage(test_storage_space, wtx)
    bc = Blockchain(test_storage_space)
    mptx = MempoolTx(test_storage_space)
    utxoi = UTXOIndex(test_storage_space, wtx)
  def f(smth):
    pass
  bc.ask_for_blocks_hook = f
  bs.ask_for_txouts_hook = f
  return test_storage_space

test_storage_space=rebuild_test_storage_space()




from leer.core.storage.merkle_storage import MMR
import shutil, os, time, lmdb

class MMRTest1(MMR):
  def sum(self,x,y):
   return bytes("(%s+%s)"%(x.decode('utf-8'),y.decode('utf-8')),'ascii')

env = None
def create_db():
  global env
  path = "~/.testleer/"
  if not os.path.exists(path): 
      os.makedirs(path) #TODO catch
  env = lmdb.open(path, map_size = 1000000, max_dbs=70)   


def wipe_test_dirs():
  shutil.rmtree("~/.testleer/", True)

def basic_test(env, wtx):
  #((((0+1)+(2+3))+((4+5)+(6+7)))+(8+9))
  #0        1        2        3        4        5        6        7        8        9        
  # \      /          \      /          \      /          \      /          \      /
  #  (0+1)             (2+3)              (4+5)            (6+7)              (8+9)
  #       \            /                  \                 /                  |
  #        ((0+1)+(2+3))                      ((4+5)+(6+7))                  (8+9)
  #              \                                  /                          |
  #                (((0+1)+(2+3))+((4+5)+(6+7)))                             (8+9)
  #                                \                                        /
  #                                   ((((0+1)+(2+3))+((4+5)+(6+7)))+(8+9))
  a=MMRTest1("test2","~/.testleer/", env, wtx)
  for i in range(10):
    a.append(wtx, bytes(str(i),'ascii'), bytes(str(i),'ascii'))

  assert a.get_root(rtx=wtx)==b'((((0+1)+(2+3))+((4+5)+(6+7)))+(8+9))'
  print("basic_test OK")


def test_rebuild_from_disc(env, wtx):
  a=MMRTest1("test1","~/.testleer/", env, wtx)
  for i in range(10):
    a.append(wtx, bytes(str(i),'ascii'),bytes(str(i),'ascii'))

  assert a.get_root(rtx=wtx)==b'((((0+1)+(2+3))+((4+5)+(6+7)))+(8+9))'
  b=MMRTest1("test1","~/.testleer/", env, wtx)
  assert b.get_root(rtx=wtx)==b'((((0+1)+(2+3))+((4+5)+(6+7)))+(8+9))'
  print("rebuild_from_disc OK")


def test_remove_elements(env, wtx):
  a=MMRTest1("test3","~/.testleer/", env, wtx)
  for i in range(10):
    a.append(wtx, bytes(str(i),'ascii'),bytes(str(i),'ascii'))

  assert a.get_root(rtx=wtx)==b'((((0+1)+(2+3))+((4+5)+(6+7)))+(8+9))'
  a.remove(1, wtx=wtx)
  assert a.get_root(rtx=wtx)==b'((((0+1)+(2+3))+((4+5)+(6+7)))+8)'

  #This whole construction is nasty, should be done with proper exceptions
  bad =False
  try: 
    a.remove(1, wtx=wtx, set_of_indexes={b'3'})
    bad = True
  except:
    pass
  if bad:
    raise

  a.append(wtx, bytes(str(9),'ascii'),bytes(str(i),'ascii'))
  assert a.get_root(rtx=wtx)==b'((((0+1)+(2+3))+((4+5)+(6+7)))+(8+9))'
  a.remove(4, wtx=wtx)
  assert a.get_root(rtx=wtx)==b'(((0+1)+(2+3))+(4+5))'
  a.append(wtx, bytes('a','ascii'),bytes(str(i),'ascii'))
  a.append(wtx, bytes('b','ascii'),bytes(str(i),'ascii'))
  a.append(wtx, bytes('c','ascii'),bytes(str(i),'ascii'))
  assert a.get_root(rtx=wtx)==b'((((0+1)+(2+3))+((4+5)+(a+b)))+c)'
  print("test_element_removal OK")



def test_prune_elements(env, wtx):
  a=MMRTest1("test4","~/.testleer/", env, wtx)
  for i in range(10):
    a.append(wtx, bytes(str(i),'ascii'),bytes(str(i)*2,'ascii'))
  full_root = b'((((0+1)+(2+3))+((4+5)+(6+7)))+(8+9))'
  assert a.get_root(rtx=wtx)==full_root
  cleared_objects=[]
  for i in range(3,6):
    cleared_objects.append(a.clear(bytes(str(i),'ascii'), wtx=wtx))
  assert a.get_root(rtx=wtx)==b'((((0+1)+2)+(6+7))+(8+9))'
  print("test_element_clearing OK")
  for _o in cleared_objects[:-1]:
    a.revert_clearing(_o, wtx=wtx)
  assert a.get_root(rtx=wtx)==b'((((0+1)+(2+3))+(4+(6+7)))+(8+9))'
  a.revert_clearing(cleared_objects[-1], wtx=wtx)
  assert a.get_root(rtx=wtx)==full_root
  print("test_element_revert_clearing OK")

  a=MMRTest1("test5","~/.testleer/", env, wtx)
  for i in range(10):
    a.append(wtx, bytes(str(i),'ascii'),bytes(str(i)*2,'ascii'))

  assert a.get_root(rtx=wtx)==full_root
  discarded_objects=[]
  for i in range(3,6):
    discarded_objects.append(a.discard(bytes(str(i),'ascii'), wtx=wtx))
  for i in range(10):
    el=a.get_by_hash(bytes(str(i),'ascii'), rtx=wtx)
    if i in list(range(3,6)):
      assert el==None
    else:
      assert el==bytes(str(i)*2,'ascii')
  assert a.get_root(rtx=wtx)==full_root
  print("test_element_discarding OK")
  for _o in discarded_objects[:-1]:
    a.revert_discarding(_o, wtx=wtx)
  assert a.get_root(rtx=wtx)==full_root
  a.revert_discarding(cleared_objects[-1], wtx=wtx)
  assert a.get_root(rtx=wtx)==full_root
  for i in range(10):
    el=a.get_by_hash(bytes(str(i),'ascii'), rtx=wtx)
    assert el==bytes(str(i)*2,'ascii')
  print("test_element_revert_discarding OK")


def test_save_pruned_db(env, wtx):
  a=MMRTest1("test6","~/.testleer/", env, wtx, save_pruned=True)
  for i in range(10):
    a.append(wtx, bytes(str(i),'ascii'),bytes(str(i)*2,'ascii'))
  full_root = b'((((0+1)+(2+3))+((4+5)+(6+7)))+(8+9))'
  assert a.get_root(rtx=wtx)==full_root
  cleared_objects=[]
  for i in range(3,6):
    cleared_objects.append(a.clear(bytes(str(i),'ascii'), wtx=wtx))
  assert a.get_root(rtx=wtx)==b'((((0+1)+2)+(6+7))+(8+9))'
  for i in range(3,6):
    assert a.get_pruned_by_hash(bytes(str(i),'ascii'),rtx=wtx)==bytes(str(i)*2,'ascii')
  print("test_pruned_in_pruned_db_clear OK")
  for _o in cleared_objects[:-1]:
    a.revert_clearing(_o, wtx=wtx)
  assert a.get_root(rtx=wtx)==b'((((0+1)+(2+3))+(4+(6+7)))+(8+9))'
  a.revert_clearing(cleared_objects[-1], wtx=wtx)
  assert a.get_root(rtx=wtx)==full_root
  for i in range(3,6):
    assert a.get_pruned_by_hash(bytes(str(i),'ascii'), rtx=wtx)==None
  print("test_pruned_not_in_pruned_db_afeter_revert_clearing OK")

  a=MMRTest1("test7","~/.testleer/", env, wtx)
  for i in range(10):
    a.append(wtx, bytes(str(i),'ascii'),bytes(str(i)*2,'ascii'))

  assert a.get_root(rtx=wtx)==full_root
  discarded_objects=[]
  for i in range(3,6):
    discarded_objects.append(a.discard(bytes(str(i),'ascii'), wtx=wtx))
  for i in range(10):
    el=a.get_by_hash(bytes(str(i),'ascii'), rtx=wtx)
    if i in list(range(3,6)):
      assert el==None
    else:
      assert el==bytes(str(i)*2,'ascii')
  assert a.get_root(rtx=wtx)==full_root
  print("test_pruned_in_pruned_db_clear OK")
  for _o in discarded_objects[:-1]:
    a.revert_discarding(_o, wtx=wtx)
  assert a.get_root(rtx=wtx)==full_root
  a.revert_discarding(cleared_objects[-1], wtx=wtx)
  assert a.get_root(rtx=wtx)==full_root
  for i in range(10):
    el=a.get_by_hash(bytes(str(i),'ascii'),rtx=wtx)
    assert el==bytes(str(i)*2,'ascii')
  print("test_element_revert_discarding OK")


def test_state_assignment(env, wtx):
  a=MMRTest1("test7","~/.testleer/", env, wtx, save_pruned=True)
  assert a.get_state(rtx=wtx)==None
  new_state = b"\x33"*32
  a.set_state(new_state, wtx=wtx)
  assert new_state == a.get_state(rtx=wtx)
  print("test_state_assignment OK")


def test_unique(env, wtx):
  a=MMRTest1("unique","~/.testleer/", env, wtx, save_pruned=True)
  for i in range(10):
    a.append_unique(wtx, bytes(str(i),'ascii'), b"e1ee7 "+bytes(str(i),'ascii'))
  assert a.get_root(rtx=wtx)==b'((((0+1)+(2+3))+((4+5)+(6+7)))+(8+9))'
  try:
    a.append_unique(wtx, bytes(str(8),'ascii'), bytes(str(8),'ascii'))
    raise Exception
  except KeyError:
    pass
  assert a.has_index(rtx=wtx, obj_index=bytes(str(8),'ascii'))
  assert not a.has_index(rtx=wtx, obj_index=bytes(str(12),'ascii'))
  a.update_index(wtx, 2, b"e")
  assert a.get_root(rtx=wtx)==b'((((0+1)+(e+3))+((4+5)+(6+7)))+(8+9))'
  assert not a.has_index(rtx=wtx, obj_index=bytes(str(2),'ascii'))
  assert a.has_index(rtx=wtx, obj_index=b"e")
  assert a.get_by_hash(b"e", rtx=wtx) == b"e1ee7 2"
  print("test_unique OK")
  



def bench(env, wtx, n=10000):
  tm=time.time()
  a=MMRTest1("test1","~/.testleer/", env, wtx)
  for i in range(n):
    a.append(wtx, bytes(str(i),'ascii'),bytes(str(i),'ascii'))

  assert a.get_root(rtx=wtx)==b"("*5
  print("Initial filling of %d take %f sec"%(n, time.time()-tm))
  
  tm=time.time()
  b=MMRTest1("test1","~/.testleer/", env, wtx)
  assert b.get_root(rtx=wtx)==b"("*5
  print("Reload take of %d take %f sec"%(n, time.time()-tm))


def merkle_test():
  wipe_test_dirs()
  create_db()
  with env.begin(write=True) as wtx:
    basic_test(env, wtx)
    test_rebuild_from_disc(env, wtx)
    test_remove_elements(env, wtx)
    test_prune_elements(env, wtx)
    test_save_pruned_db(env, wtx)
    test_state_assignment(env, wtx)
    test_unique(env, wtx)
  wipe_test_dirs()




import asyncio
import logging
import sys
from concurrent.futures import CancelledError

from secp256k1_zkp import PrivateKey, PublicKey

from lnoise import Key, HandshakeState, PartialData
from .messages import inv_message_id

async def tripple_read(r):
    message = await r(65536)
    if not len(message):
      message = await r(65536)
    if not len(message):
      message = await r(65536)
    return message


class NetworkNode:
  def __init__(self, our_node, params, loop, serialized_params=None):
    if serialized_params:
      params = self.deserialize_params(serialized_params)
    self.host = params['network']['host']
    self.port = params['network']['port']
    self.advertised_host = params['network'].get('advertised_host', self.host)
    self.advertised_port = params['network'].get('advertised_port', self.port)
    self.static_key = params.get('static_key', None)
    self.advertised_static_key = params.get('advertised_static_key', self.static_key)
    self.static_full_key = params.get('static_full_key', None) #This option is only for our node
    if self.static_full_key and not self.static_key:
      self.static_key = self.static_full_key.key.pubkey
    self.ephemeral_key = Key() #generating new key on the fly
    self.our_node = our_node
    self.loop = loop
    self.listening_loop_task = None
    self.connected= False
    self.reader, self.writer = None, None
    self.handshake=HandshakeState()
    if self.our_node:
      self.logger=logging.getLogger(__name__+" Node(%s,%s from (%s))"%(str(self.host),str(self.port),str(self.our_node.port)))
    else:
      self.logger=logging.getLogger(__name__+" Node(%s,%s our_node)"%(str(self.host),str(self.port)))
    self.partial_data = None

  def serialize_params(self):
    ser_host = str(self.advertised_host).encode("utf-8")
    ser_port = str(self.advertised_port).encode("utf-8")
    ser_static_key = self.static_key.serialize()
    result = b""
    return len(ser_host).to_bytes(1, "big") + ser_host +len(ser_port).to_bytes(1, "big") + ser_port +len(ser_static_key).to_bytes(1, "big") + ser_static_key

  def deserialize_params(self, serialized_params, remote=False):
    try:
      len_ser_host, r = serialized_params[0], serialized_params[1:]
      ser_host, r = r[0:len_ser_host], r[len_ser_host:]
      len_ser_port, r = r[0], r[1:]
      ser_port, r = r[0:len_ser_port], r[len_ser_port:]
      len_ser_key, r = r[0], r[1:]
      ser_key, r = r[0:len_ser_key], r[len_ser_key:]
    except:
      raise Exception("Incorrect serialized params") #TODO handle exceptions more precisely
    if remote:
      self.advertised_host = ser_host.decode('utf-8')
      self.advertised_port = int(ser_port.decode('utf-8')) # TODO try-except here
      self.advertised_static_key = PublicKey(bytes(ser_key), raw=True)
    else:
      return {'network':{'host':ser_host.decode('utf-8'), 'port':int(ser_port.decode('utf-8'))}, 'static_key':PublicKey(bytes(ser_key), raw=True)}

  async def connect(self):
    '''
      Note: to open connection we already should know remote static key 
    '''
    self.logger.debug("Trying to connect")
    if self.connected:
      raise
    try:
      self.reader, self.writer = await asyncio.open_connection(self.host, self.port, loop=self.loop)
    except (ConnectionRefusedError,TimeoutError, OSError) as e:
      # TODO somehow instead of ConnectionRefused, OSError(Multiple exceptions) happens
      # should be cleared
      self.logger.info("Node (%s:%s) is not reachable `%s`"%(self.host, self.port, str(e)))
      self._on_closed_connection()
      return
    except Exception as e:
      self._on_closed_connection(error=e)
      return
    self.handshake.initialize('Noise_XK', prologue=b'leer', s=self.our_node.static_full_key, e=self.our_node.ephemeral_key, rs=self.static_key, re=None, initiator=True)
    try:
      await self.do_handshake()
    except Exception as e:
      exc_type, exc_value, exc_traceback = sys.exc_info()
      self.logger.info("Close connection due to error in handshake `%s`"%(str(e)), exc_info=True)
      self._on_closed_connection(error=e)
    

  async def accept_connection(self, reader, writer):
    if self.connected:
      raise
    self.logger.debug("Trying to accept connection")
    self.reader, self.writer = reader, writer
    if self.reader.at_eof():
      return
    self.handshake.initialize('Noise_XK', prologue=b'leer', s=self.our_node.static_full_key, e=self.our_node.ephemeral_key, rs=None, re=None, initiator=False)
    #self.connected = True
    try:
      await self.do_handshake()
    except Exception as e:
      exc_type, exc_value, exc_traceback = sys.exc_info()
      self.logger.info("Close connection due to error in handshake `%s`"%(str(e)), exc_info=True)
      self._on_closed_connection(error=e)

  async def do_handshake(self):
    if not self.handshake.state in [self.handshake.STATES.WAITING_TO_WRITE, self.handshake.STATES.WAITING_TO_READ]:#wait to read or write
      raise
    elif self.handshake.state==self.handshake.STATES.WAITING_TO_WRITE:
        message = self.handshake.write_message(b"")
        self.writer.write(message)
        await self.writer.drain()
    elif self.handshake.state==self.handshake.STATES.WAITING_TO_READ:
        mb=[]
        message = await tripple_read(self.reader.read)
        self.handshake.read_message(message, [])
    if not self.handshake.state==self.handshake.STATES.ESTABLISHED:
      await self.do_handshake()
    else:
      self.static_key = self.handshake.rs
      self.advertised_static_key = self.static_key #can be overwritten by `init` message
      self.listening_loop_task = asyncio.ensure_future(self.listening_loop())
      asyncio.ensure_future(self.on_established_connection())
      self.connected = True

  async def listening_loop(self):
    try:
      while self.connected:
        if not self.handshake.state==self.handshake.STATES.ESTABLISHED:
            await asyncio.sleep(1)
            continue
        data = await self.reader.read(65536) 
        if (not data) or self.writer.transport.is_closing():
          self._on_closed_connection()
          break
        else:
            try:
                data = self.partial_data + data if self.partial_data else data              
                self.partial_data = None
                while len(data):
                  message, data = self.handshake.session.decode(data)
                  if message in [inv_message_id["ping"], inv_message_id["pong"]]:
                    self.logger.debug("Got message %s"%(str(message)))
                  else:
                    if len(message)<24:
                      self.logger.info("Got message %s"%(str(message)))
                    else:
                      self.logger.info("Got message with len %d"%(len(message)))
                      self.logger.debug("Message content %s"%(str(message)))
                  result = await self.handle_message(message)
            except PartialData:
              self.partial_data=data
            except CancelledError:
              return
            except Exception as e:
                self._on_closed_connection(error=e)
    except CancelledError:
      pass

  async def send(self,message):
    try:
      message = message
      while not (self.handshake.state==self.handshake.STATES.ESTABLISHED):
        await asyncio.sleep(1)
      while self.writer._protocol._drain_waiter:
        await asyncio.sleep(0.03) #wait for previous send to complete
      encoded_message = self.handshake.session.encode(message)
      self.writer._transport.set_write_buffer_limits(high=0)
      self.writer.write(encoded_message)
      await self.writer.drain()
    except ConnectionResetError as e:
      self._on_closed_connection(error=e)
      return False
    except Exception as e:
      self.logger.info("Exception during send")
      self._on_closed_connection(error=e)
      return False
    return True

  async def on_established_connection(self):
    '''
      This method is expected to be overwritten by subclass
    '''
    pass

  def _on_closed_connection(self, error=None):
    self.connected = False
    if not error:
      self.logger.info("Connection is closed, stop listening_loop")
    else:
      self.logger.info("Close connection due to error: `%s`"%(str(error)))
    if self.listening_loop_task:
      self.listening_loop_task.cancel()
    if self.writer:
      if not self.writer.transport.is_closing():
        self.writer.close()
    self.handshake=HandshakeState()
    asyncio.ensure_future(self.on_closed_connection())

  async def on_closed_connection(self):
    '''
      This method is expected to be overwritten by subclass
    '''
    pass

  async def handle_message(self, message):
    '''
      This method is expected to be overwritten by subclass
    '''
    pass



from leer.core.primitives.header import Header, ContextHeader
import shutil, os, time, lmdb, math

class HeadersStorage:

  __shared_states = {}

  def __init__(self, storage_space, wtx):
    path = storage_space.path
    if not path in self.__shared_states:
        self.__shared_states[path]={}
    self.__dict__ = self.__shared_states[path]
    self.storage = HeadersDiscStorage(path, env=storage_space.env, wtx=wtx)
    self.storage_space = storage_space
    self.storage_space.register_headers_storage(self)

  def get(self, _hash, rtx):
    serialized_header = self.storage.get_by_hash(_hash, rtx=rtx)
    if not serialized_header:
      raise KeyError(_hash)
    ch=ContextHeader()
    ch.deserialize(serialized_header)
    return ch

  def put(self, _hash, header, wtx):
    self.storage.put(header.height, header.hash, header.serialize_with_context(), wtx=wtx)

  def update(self, _hash, header, wtx):
    self.storage.update(header.hash, header.serialize_with_context(), wtx=wtx)

  def has(self, _hash, rtx):
    return bool(self.storage.get_by_hash(_hash, rtx=rtx))
      
  def get_headers_at_height(self, height, rtx):
    ret=[]
    for serialized_header in self.storage.get_by_height(height, rtx=rtx):
      ch=ContextHeader()
      ch.deserialize(serialized_header)
      ret.append(ch)
    return ret

  def get_headers_hashes_at_height(self, height, rtx):
    return self.storage.get_hashes_by_height(height,rtx=rtx)


def __(x):
  return (x).to_bytes(4,'big')

class HeadersDiscStorage:
  def __init__(self, dir_path, env, wtx):
    self.dir_path = dir_path

    self.env = env
    self.main_db = self.env.open_db(b'headers_main_db', txn=wtx, dupsort=False)
    self.height_db = self.env.open_db(b'headers_height_db', txn=wtx, dupsort=True)

  def put(self, height, _hash, serialized_header, wtx):
    p1=wtx.put( bytes(_hash), bytes(serialized_header), db=self.main_db, dupdata=False, overwrite=True)
    p2=wtx.put( __(height), bytes(_hash), db=self.height_db, dupdata=True)

  def update(self, _hash, serialized_header, wtx):
    wtx.put(bytes(_hash), bytes(serialized_header), db=self.main_db, dupdata=False, overwrite=True)

  def get_by_hash(self, _hash, rtx):
    return rtx.get(bytes(_hash), db=self.main_db)

  def get_by_height(self, height, rtx):
      cursor = rtx.cursor(db=self.height_db)
      assert cursor.set_key(__(height))
      _hashes = list(cursor.iternext_dup())
      return [self.get_by_hash(_hash, rtx=rtx) for _hash in _hashes] 
  
  def get_hashes_by_height(self, height, rtx):
      cursor = rtx.cursor(db=self.height_db)
      assert cursor.set_key(__(height))
      return list(cursor.iternext_dup())
  




import hashlib
from leer.core.utils import encode_target, decode_target
from leer.core.hash.progpow import progpow_hash, partial_hash

class PoPoW:
  # We use compact version of PoPoW (https://eprint.iacr.org/2017/963.pdf) here
  # Differences from the aforementioned protocol are:
  # 1. \mu-level chain is chain of headers with hashes (4+2*\mu) zeroes (less
  #    than 2**(256-(4+\mu)*4) )
  # 2. All interlinks with level less than previous header level are skipped
  #    (they are all equal prev hash anyway)
  # 3. If \mu-level chain is empty (there was no header with so much work in it)
  #    this level is not included. However, genesis header hash is always last
  #    in the list. Maximum level is not fixed and thus PoPoW has variable length 
  #    and should contain length. Genesis header is appended to the list to 
  #    prevent using PoPoW from another chain

  def __init__(self, pointers=None):
    self.pointers = pointers
    
  @property
  def prev(self):
    if not len(self.pointers):
      #Genesis header has empty popow
      return b"\x00"*32
    return self.pointers[0]

  def serialize(self):
    ser = len(self.pointers).to_bytes(1,'big')
    for pointer in self.pointers:
      ser += pointer
    return ser

  def deserialize(self, serialized_popow):
    self.deserialize_raw(serialized_popow)

  def deserialize_raw(self, serialized_popow):
    # This function derserializes popow and returns unused serialized data
    self.pointers=[]
    if len(serialized_popow)<1:
      raise Exception("Not enough bytes in PoPoW to store length")
    _len, serialized_popow = serialized_popow[0], serialized_popow[1:]
    for i in range(_len):
      if len(serialized_popow)<32:
        raise Exception("Not enough bytes in PoPoW to store %d pointer"%i)
      pointer,serialized_popow = serialized_popow[:32], serialized_popow[32:]
      self.pointers.append(pointer)
    return serialized_popow

  def check_self_consistency(self):
    # 1. Check that all levels are consistent, except last one (genesis)
    if not len(self.pointers):
      #Genesis block
      return True
    level = self._get_level(self.pointers[0])
    for i in self.pointers[1:-1]:
      next_level = self._get_level(i)
      if level>=next_level:
        return False
    return True

  def _get_level(self, _hash):
    level=-2
    for i in _hash: # Python3: getting element of bytes return integer
      if not i==0:
        break
      level+=1
    return level

  def generate_from_prev(self, prev_popow, prev_hash):
    prev_level = self._get_level(prev_hash)
    self.pointers = [prev_hash]
    current_level = prev_level
    for i in prev_popow.pointers:
      level = self._get_level(i)
      if level<=current_level or level<0:
        continue
      else:
        self.pointers.append(i)
        current_level = self._get_level(i)
    genesis = prev_popow.pointers[-1] if len(prev_popow.pointers) else prev_hash
    if not self.pointers[-1]==genesis:
      self.pointers.append(genesis)
    
  def __eq__(self, another):
    return self.pointers==another.pointers



class VoteData:
  # VoteData contains vector for voting (something like https://github.com/bitcoin/bips/blob/master/bip-0009.mediawiki)
  # This vector is 4 bytes long: 1 byte for hard forks and 3 bytes for soft forks
  # Also it contains one specific type of voting: voting for miner susbsidy which is planned to be variable (under constraints)
  # MinerSubsidyVote is 1 byte long and represents setting reward from 0 to 100% of maximal at this height

  # TODO defaults should be set at higher level
  def __init__(self, forks_vector=b"\x00"*4, miner_subsidy_vote=b"\xff"):
    self.forks_vector = forks_vector
    self.miner_subsidy_vote = miner_subsidy_vote

  @property
  def miner_subsidy_vote_int(self):
    return int.from_bytes(self.miner_subsidy_vote, "big")

  def serialize(self):
    return self.forks_vector+self.miner_subsidy_vote

  def deserialize(self, serialized):
    self.deserialize_raw(serialized)

  def deserialize_raw(self, serialized):
    if len(serialized)<5:
      raise Exception("Not enough bytes for deserialization")
    self.forks_vector, self.miner_subsidy_vote, serialized = serialized[:4], serialized[4:5], serialized[5:]
    return serialized


class Header:
  def __init__(self, height=None, supply=None, full_offset = None, merkles=None, popow=None, votedata=None, timestamp=None, target=None, version=None, nonce=None):
    self.height = height
    self.version = version
    self.supply = supply #since miners can vote for changing reward supply is not predetermined function of height
    self.full_offset = full_offset # sum of offsets of all transactions
    self.merkles = merkles
    self.popow = popow
    self.votedata = votedata
    self.timestamp = timestamp
    self.target = target
    self.extension_bytes = b"\x00\x00\x00\x00"
    self.nonce = nonce

  @property
  def serialized_merkles(self):
    # self.merkles is list of 3 merkle roots:
    # 1. (unspent) outputs commitment
    # 2. (unspent) outputs data (rangeproof and so on)
    # 3. (unmerged) excesses
    assert len(self.merkles[0])==65
    assert len(self.merkles[1])==32
    assert len(self.merkles[2])==65
    return self.merkles[0] + self.merkles[1] + self.merkles[2]

  @property
  def encoded_target(self):
    significand, order = encode_target(self.target)
    return significand.to_bytes(1,"big") + order.to_bytes(1,"big")

  @property
  def difficulty(self):
    return (2**256)//self.target

  @property
  def template(self):
    '''
      Serialized header without nonce. Templates are used during mining, when we search for acceptable nonce
    '''
    return self.version.to_bytes(1, "big") +\
           self.height.to_bytes(4,"big") +\
           self.popow.serialize() +\
           self.votedata.serialize() + \
           self.serialized_merkles + \
           self.supply.to_bytes(8,"big") + \
           self.full_offset.to_bytes(32,"big") + \
           int(self.timestamp).to_bytes(5, "big") +\
           self.encoded_target +\
           self.extension_bytes

  def serialize(self):
    return self.template + self.nonce

  def deserialize(self, serialized):
    self.deserialize_raw(serialized)

  def deserialize_raw(self, serialized):
    if len(serialized)<1:
      raise Exception("Not enough bytes for version deserialization")
    self.version, serialized = serialized[0], serialized[1:]
    if len(serialized)<4:
      raise Exception("Not enough bytes for height deserialization")
    self.height, serialized = int.from_bytes(serialized[:4], "big"), serialized[4:]
    self.popow = PoPoW()
    serialized = self.popow.deserialize_raw(serialized)
    self.votedata = VoteData()
    serialized = self.votedata.deserialize_raw(serialized)
    if len(serialized)<162:
      raise Exception("Not enough bytes for merkle roots deserialization")
    self.merkles, serialized = [serialized[:65],serialized[65:97],serialized[97:162]], serialized[162:]
    if len(serialized)<8:
      raise Exception("Not enough bytes for supply deserialization")
    self.supply, serialized = int.from_bytes(serialized[:8], "big"), serialized[8:]
    if len(serialized)<32:
      raise Exception("Not enough bytes for full_offset deserialization")
    self.full_offset, serialized = int.from_bytes(serialized[:32], "big"), serialized[32:]
    if len(serialized)<5:
      raise Exception("Not enough bytes for timestamp deserialization")
    self.timestamp, serialized = int.from_bytes(serialized[:5], "big"), serialized[5:]
    if len(serialized)<5:
      raise Exception("Not enough bytes for target deserialization")
    self.target, serialized = decode_target(serialized[0],serialized[1]), serialized[2:]
    if len(serialized)<4:
      raise Exception("Not enough bytes for extension bytes deserialization")
    self.extension_bytes, serialized = serialized[:4], serialized[4:]
    if len(serialized)<8:
      raise Exception("Not enough bytes for nonce deserialization")
    self.nonce, serialized = serialized[:8], serialized[8:]
    return serialized
    

  def check_self_consistency(self):
    popow_check = self.popow.check_self_consistency()
    # TODO header by itself has enough information to check wether supply and merkles consistent
    return popow_check

  def score(self):
    pass

  @property
  def hash(self):
    return progpow_hash(self.height, self.template, self.nonce)

  @property
  def partial_hash(self):
    return partial_hash(self.template)

  @property
  def integer_hash(self):
    return int.from_bytes(self.hash, "big")

  @property
  def prev(self):
    return self.popow.prev

  def next_popow(self):
    template = PoPoW()
    template.generate_from_prev(self.popow, self.hash)
    return template

  def __repr__(self):
    return "<Header %s>"%self.hash

  def __eq__(self, h):
    return (self.height == h.height and
            self.popow == h.popow and
            self.votedata.serialize() == h.votedata.serialize() and
            self.merkles == h.merkles and
            self.supply == h.supply and
            self.timestamp == h.timestamp and
            self.encoded_target == h.encoded_target and
            self.version == h.version and
            self.nonce == h.nonce)


class ContextHeader(Header):
  def __init__(self, header=None):
    if header:
      Header.__init__(self, height = header.height,
                            supply = header.supply,
                            full_offset = header.full_offset,
                            merkles = header.merkles,
                            popow = header.popow,
                            votedata = header.votedata,
                            timestamp = header.timestamp,
                            target = header.target,
                            version = header.version,
                            nonce = header.nonce)
    else:
      Header.__init__(self)
    self.descendants = set()
    self.connected_to_genesis = False
    self.invalid = False
    self.reason = None
    '''
      difference between coins_to_be_mint and supply is sum of `new_outputs_fee`s.
      We set (by default) coins_to_be_mint to 0. It will be overwritten when header becomes connected to genesis
      and value will be required during context validation.
    '''
    self.coins_to_be_mint = 0 
    self.total_difficulty = 0 

  def serialize_with_context(self):
    ser = super(ContextHeader, self).serialize()
    #TODO what if attacker creates more than 255 headers?
    #It may be extremely cheap for old blocks
    ser += len(self.descendants).to_bytes(1,'big')
    for descendant in self.descendants:
      ser += descendant
    ser += int(self.connected_to_genesis).to_bytes(1,'big')
    ser += int(self.invalid).to_bytes(1,'big')
    reason = self.reason if self.reason else ""
    ser += int(len(reason)).to_bytes(2,'big')
    ser += reason.encode('utf-8')
    ser += self.coins_to_be_mint.to_bytes(8,"big")
    ser += self.total_difficulty.to_bytes(32,"big")
    return ser

  def deserialize(self, serialized):
    self.deserialize_raw(serialized)

  def deserialize_raw(self, serialized):
    #TODO exceptions for not enough bytes
    # No urgency: we never should get contextHeader from other nodes
    ser = super(ContextHeader, self).deserialize_raw(serialized)
    desc_num, ser = int.from_bytes(ser[:1], 'big'), ser[1:]
    for desc in range(desc_num):
      d, ser = ser[:32], ser[32:]
      self.descendants.add(d)
    self.connected_to_genesis, self.invalid, ser = bool(ser[0]), bool(ser[1]), ser[2:]
    reason_len, ser = int.from_bytes(ser[:2], 'big'), ser[2:]
    self.reason, ser = ser[:reason_len].decode('utf-8'), ser[reason_len:]
    self.coins_to_be_mint, ser = int.from_bytes(ser[:8], "big"), ser[8:]
    self.total_difficulty, ser = int.from_bytes(ser[:32], "big"), ser[32:]
    return ser
    






import functools
from leer.core.lubbadubdub.offset_utils import subtract_offset
from chacha20poly1305 import ChaCha20Poly1305
from secp256k1_zkp import PrivateKey, PublicKey, PedersenCommitment, default_blinding_generator
from leer.core.lubbadubdub.constants import default_generator

def top_up_nonce(nonce):
    if len(nonce)<12:
      return b"\x00"*(12-len(nonce))+nonce
    else:
      return nonce[-12:]


def encrypt(pubkey, nonce, plaintext):
    '''
      Encrypt plaintext for some pubkey (owner of corresponding privkey can decrypt).
      params:
       pubkey: secp256k1_py PublicKey object
       nonce: 12 bytes nonce, should not be reused
       plaintext: bytes object with arbitrary length.
      
      Inner logic:
        1) generate ephemereal (one time) private_key.
        2) generate shared_secret : ecdh of private_key and receiver's pubkey
        3) symmetrically encrypt plaintext with shared_secret (encryption with ChaCha20Poly1305)
        4) attach ephemereal public_key to ciphertext
    '''
    nonce = top_up_nonce(nonce)
    ephemeral_privkey = PrivateKey()
    shared_key = pubkey.ecdh(ephemeral_privkey.private_key)
    aead = ChaCha20Poly1305(shared_key, 'python')
    res = aead.seal(nonce, plaintext, b'')
    res = ephemeral_privkey.pubkey.serialize() + res
    return res


def decrypt(privkey, nonce, ciphertext):
    nonce = top_up_nonce(nonce)
    pubkey, ciphertext = PublicKey(pubkey=ciphertext[:33], raw=True), ciphertext[33:]
    shared_key = pubkey.ecdh(privkey.private_key)
    aead = ChaCha20Poly1305(shared_key, 'python')
    res = aead.open(nonce, ciphertext, b'')
    if res is None:
      raise Exception("Cant decrypt")
    return res

@functools.lru_cache(maxsize=30)
def compare_supply_and_merkle_roots(total_supply, commitment_root, excesses_root, full_offset):
  '''
    Each txout (authorized pedersen) commitment is v*H + r*G, where v is value and r is blinding key, G and H - generators.
    Each blinding key is previous blinding key + private_key of address +/- blinding key which should be compensated by additional excesses.
    Thus if we summarized all commitments result should V*H+R*G, where
    V - is supply: sum of all unspend values on blockchain. Note that supply is not equal to all minted coins, 
        since new_outputs_fee retains some coins.
    R - is summ of private keys, private keys of additional excsses and full_offset (summ of offset of all transactions). Thus R*H is summ of all addresses's pubkeys, all additional excsses and full_offset*H. Note, that while R is not known, R*G can be calculated from public data.
    This function check that calculated
        V*H (calculated from supply) plus R*G (calculated from excesses and addresses) + full_offset*G is equal to summ of commitments V*H+R*G.
  '''
  commitment_summ = PedersenCommitment(commitment=commitment_root[:33], raw=True) 
  excesses_summ = PublicKey(pubkey= excesses_root[:33], raw=True).to_pedersen_commitment()
  # Instead of generating separately supply_pc (which is actually public key v*H) and
  # full_offset_pc (which is actually public key fo*G) lets generate sum
  minus_fo = subtract_offset(0, full_offset)
  supply_and_offset_pc = PedersenCommitment(value_generator = default_generator, blinding_generator = default_blinding_generator)
  supply_and_offset_pc.create(total_supply, minus_fo.to_bytes(32,"big"))
  checker = PedersenCommitment()
  return checker.verify_sum([excesses_summ, supply_and_offset_pc], [commitment_summ])
  





import os, lmdb

class StorageSpace:
  '''
    Storage space is a set of objects representing blockchain.
    Two different storage spaces, for instance testnet and mainnet should not
    share any objects.
    In the future it will be literally a combination of all storages in one physical storage,
    thus truly atomic updates will be possible
  '''
  def __init__(self, path):
    self.path = path
    if not os.path.exists(path): 
        os.makedirs(self.path) #TODO catch
    _25GB = int(25 * 1e9)
    self.env = lmdb.open(self.path, map_size = _25GB, max_dbs=70)        

  def register_txos_storage(self, txos):
    self.txos_storage = txos

  def register_excesses_storage(self, excesses):
    self.excesses_storage = excesses

  def register_headers_storage(self, headers):
    self.headers_storage = headers

  def register_headers_manager(self, headers_manager):
    self.headers_manager = headers_manager

  def register_blocks_storage(self, blocks):
    self.blocks_storage = blocks

  def register_blockchain(self, blockchain):
    self.blockchain = blockchain

  def register_mempool_tx(self, mempool_tx):
    self.mempool_tx = mempool_tx

  def register_utxo_index(self, utxo_index):
    self.utxo_index = utxo_index

  def get_block_template(self):
    '''
      For now blocks are space dependent
    '''
    pass




from secp256k1_zkp import PrivateKey
from hashlib import sha256

def partial_hash(_bytes):
    seed1 = _bytes
    m1 = sha256()
    m1.update(seed1)
    return m1.digest()

def mining_canary_hash_part(_bytes):
    m2 = sha256()
    m2.update(_bytes)    
    p = PrivateKey(privkey=m2.digest(), raw=True)
    m3 = sha256()
    m3.update(p.pubkey.serialize()[1:])
    return m3.digest()

def mining_canary_hash(header_height, serialized_header_without_nonce, nonce_bytes):
    ph = partial_hash(serialized_header_without_nonce)
    return mining_canary_hash_part(ph+nonce_bytes)




default_config =   {
                     'relay_fee_per_kb': 3000
                   }


class FeePolicyChecker:
  
  def __init__(self, config):
    self.relay_fee_per_kb = int(config['relay_fee_per_kb'])

  def check_tx(self,tx):
    size = len(tx.serialize())
    if tx.coinbase:
      size -= 2 #len of len of output
      size -= len(tx.coinbase.serialize())
    if int((size/1000.)*self.relay_fee_per_kb)>tx.relay_fee:
      return False
    return True
    




from secp256k1_zkp import PrivateKey
import shutil, os, time, lmdb, math
from hashlib import sha256
from leer.core.lubbadubdub.address import address_from_private_key
import base64

class KeyManagerClass:
  '''
    KeyManager holds information about private keys, spent and unspent outputs.
    TODO: separate key storage and transaction info handling
  '''
  def __init__(self, path, password=None):
    #self.privkey = None
    #if password:
    #  password_bytes = bytes(password+")dod8q=(3fnS#904ff", "utf-16")
    #  m = sha256()
    #  m.update(password_bytes)
    #  self.privkey = Privkey(m.digest(), raw=True)
    self.wallet = DiscWallet(dir_path=path)
    self.disc_txn = self.wallet.env.begin

  def new_address(self, w_txn):
    prk=PrivateKey()
    self.wallet.add_privkey_to_pool(prk.pubkey.serialize(), prk.private_key, w_txn)
    privkey = PrivateKey(self.wallet.get_privkey_from_pool(w_txn), raw=True)
    return address_from_private_key(privkey) 

  def priv_by_pub(self, pubkey, r_txn):
    raw_priv = self.wallet.get_privkey(pubkey.serialize(), r_txn)
    if not raw_priv:
      raise KeyError("Private key not in the wallet")
    return PrivateKey(raw_priv, raw=True)

  def priv_and_pub_by_output_index(self, output_index, r_txn):
       return self.wallet.get_priv_and_pub_by_output_index(output_index, r_txn)

  def get_output_private_data(self, output_index, r_txn):
     return self.wallet.get_output_private_data(output_index, r_txn)

  def priv_by_address(self, address, r_txn):
    raw_priv = self.wallet.get_privkey(address.pubkey.serialize(), r_txn)
    if not raw_priv:
      raise KeyError("Private key not in the wallet")
    return PrivateKey(raw_priv, raw=True)

  def add_privkey(self, privkey, w_txn):
    self.wallet.put(privkey.pubkey.serialize(), privkey.private_key, w_txn)

  def fill_pool(self, w_txn, keys_number=100):
    for _ in range(keys_number):
     prk=PrivateKey()
     self.wallet.add_privkey_to_pool(prk.pubkey.serialize(), prk.private_key, w_txn)

  #@property
  #def pool_size(self):
  #  raw_pool_size = self.wallet.get_pool_size()
  #  if raw_pool_size:
  #    return int.from_bytes(self.wallet.get_pool_size(), "big") 
  #  else: #NoneType
  #    return 0

  def is_unspent(self, output_index, r_txn):
    '''
      Note it is not check whether output is unspent or not, we check that output is marked as our and unspent in our wallet
    '''
    try:
      self.wallet.get_output(output_index, r_txn)
      return True
    except KeyError:
      return False

  def is_owned_pubkey(self, serialized_pubkey, r_txn):
    pk = self.wallet.get_privkey(serialized_pubkey, r_txn)
    if not pk:
      return False
    return True

  def spend_output(self, index, spend_height, w_txn):
    self.wallet.spend_output(index,spend_height, w_txn)

  def add_output(self, output, block_height, w_txn):
    index = output.serialized_index
    pubkey = output.address.serialized_pubkey
    taddress = output.address.to_text().encode()
    output.detect_value(inputs_info = 
         {'priv_by_pub':{
                           pubkey : self.priv_by_address(output.address, r_txn=w_txn)
                        }
         }) 
    value = output.value
    self.wallet.put_output(index, (block_height, output.lock_height, value,
                                   output.blinding_key.private_key,
                                   output.serialized_apc, taddress), w_txn)
    self.wallet.put_pubkey_output_assoc(pubkey, index, w_txn)

  def rollback(self, block_height, w_txn):
    self.wallet.remove_all_actions_in_last_block(block_height, w_txn)
    self.wallet.remove_all_outputs_created_in_block(block_height, w_txn)
    self.wallet.restore_all_outputs_spent_in_block(block_height, w_txn)

  def get_confirmed_balance_stats(self, current_height, r_txn):
    stats = {
              'matured': {'known_value':0, 'known_count':0, 'unknown_count':0},
              'immatured': {'known_value':0, 'known_count':0, 'unknown_count':0}
            }
    cursor = r_txn.cursor(db=self.wallet.pubkey_index)
    for pubkey, output_index  in cursor.iternext(values=True):
        try:
          created_height, lock_height, value, bk, apc, taddress = self.wallet.get_output(output_index, r_txn)
        except KeyError:
          continue #It's ok, output is already spent
        mat = None
        if current_height>=lock_height:
          mat = 'matured'
        else:
          mat = 'immatured'
        if value:
          stats[mat]['known_value']+=value
          stats[mat]['known_count']+=1
        else:
          stats[mat]['unknown_count']+=1        
    return stats

    
  def get_confirmed_balance_list(self, current_height, r_txn):
    ret = {}
    cursor = r_txn.cursor(db=self.wallet.pubkey_index)
    for pubkey, output_index in cursor.iternext(values=True):
        try:
          created_height, lock_height, value, bk, apc, taddress = self.wallet.get_output(output_index, r_txn)
        except KeyError:
          continue #It's ok, output is already spent
        taddress = taddress.decode()
        if not current_height>=lock_height:
          continue
        texted_index = base64.b64encode(output_index).decode()
        if not taddress in ret:
            ret[taddress]={}
        if value:
            ret[taddress][texted_index]=value
        else:
            ret[taddress][texted_index]='unknown'      
    return ret

  def give_transactions(self, n, r_txn):
    return self.wallet.generate_transaction_dict(n=n, r_txn=r_txn)

  def save_generated_transaction(self, tx, w_txn):
    '''
      Save transaction generated by wallet, thus then this tx will be found in block
      wallet will note outgoing payment.
    '''
    for o in tx.outputs:
      output_index = o.serialized_index
      params = \
        0, o.lock_height, o.value, o.blinding_key.private_key,\
        o.serialized_apc, o.address.to_text().encode()
      self.wallet.put_outgoing_output(output_index, params, w_txn)

  def is_saved(self, output, r_txn):
    try:
      self.wallet.get_outgoing_output(output.serialized_index, r_txn)
      return True
    except KeyError:
      return False

  def register_processed_output(self, output_index, block_height, w_txn):
    self.wallet.register_processed_output(output_index, w_txn)
    params = list(self.wallet.get_outgoing_output(output_index, r_txn=w_txn))
    params[0] = block_height
    self.wallet.put_outgoing_output(output_index, params, w_txn)
    

def _(x):
  return x.to_bytes(4,"big")

def _d(x):
  return int.from_bytes(x, "big")


def serialize_output_params(p):
  created_height, lock_height, value, blinding_key, serialized_apc, taddress = p
  ser_created_height = _(created_height)
  ser_lock_height = _(lock_height)
  if value == None:
    ser_value = b"\xff"*7
  else:
    ser_value = value.to_bytes(7,"big")
  return ser_created_height + ser_lock_height+ser_value+blinding_key+serialized_apc+taddress

def deserialize_output_params(p):  
  created_height, p = _d(p[:4]), p[4:]
  lock_height, p = _d(p[:4]), p[4:]
  value, p = _d(p[:7]), p[7:]
  ser_blinding_key, p = p[:32], p[32:]
  serialized_apc, p = p[:33], p[33:]
  taddress = p
  if value == 72057594037927935: #=b"\xff"*7
    value = None
  return created_height, lock_height, value, ser_blinding_key, serialized_apc, taddress


def serialize_spent_output_params(p):
  # While it is not necessary to store lock_height and created_height for spent outputs,
  # it is useful for effective unspending
  spend_height, created_height, lock_height, value, blinding_key, serialized_apc, taddress = p
  ser_spend_height = _(spend_height)
  ser_created_height = _(created_height)
  ser_lock_height = _(lock_height)
  if value == None:
    ser_value = b"\xff"*7
  else:
    ser_value = value.to_bytes(7,"big")
  return ser_spend_height+ser_created_height+ser_lock_height+ser_value+serialized_apc+blinding_key+taddress

def deserialize_spent_output_params(p):
  spend_height, p = _d(p[:4]), p[4:]
  created_height, p = _d(p[:4]), p[4:]
  lock_height, p = _d(p[:4]), p[4:]
  value, p = _d(p[:7]), p[7:]
  ser_blinding_key, p = p[:32], p[32:]
  serialized_apc, p = p[:33], p[33:]
  taddress = p
  if value == 72057594037927935: #=b"\xff"*7
    value = None
  return spend_height, created_height, lock_height, value, ser_blinding_key, serialized_apc, taddress

def repack_ser_output_to_spent(ser_output, height):
  '''
    Function for fast repacking without deserialization
  '''
  return _(height)+ser_output

def repack_ser_spent_output_to_unspent(ser_spent_output):
  '''
    Function for fast repacking without deserialization
  '''
  return ser_spent_output[4:]


RECEIVED, SPENT, SENT = b"\x00", b"\x01", b"\x02"

class DiscWallet:
  '''
    It is generally key-value db:
     1) private keys:  key is serialized pubkey; value - serialized privkey.
     2) unspent parsed outputs: key is output_index; value - tuple (lock_height, value)
     3) spent outputs: key is output_index; value - tuple (spend_height, value)
     4) block-outputs map: key is block_number; value - tuple (spent/created, output_index)
     5) pubkey-outputs map 
     6) num_index -> output: list of touched (created/spent) outputs
     7) generated outputs: key is output_index; value - output params
    There is privkey pool: bunch of pregenerated privkeys. It is expected that on a higher level
    instead of generating and immediate usage of new key, new key will be put into the pool and the oldest key
    from the pool will be used. Thus, in case of backups, copies and so on, "old copy" will contain
    some keys used after copy being made.
  '''
  def __init__(self, dir_path):
    self.dir_path = dir_path

    if not os.path.exists(self.dir_path): 
        os.makedirs(self.dir_path) #TODO catch
    self.env = lmdb.open(self.dir_path, max_dbs=10, map_size=int(250e6)) #TODO this database is too big, should be checked
    with self.env.begin(write=True) as txn:
      self.main_db = self.env.open_db(b'main_db', txn=txn, dupsort=False)
      self.pool = self.env.open_db(b'pool', txn=txn, dupsort=False)
      self.output = self.env.open_db(b'output', txn=txn, dupsort=False)
      self.spent = self.env.open_db(b'spent', txn=txn, dupsort=False)
      self.generated_outputs = self.env.open_db(b'generated', txn=txn, dupsort=False)
      self.block_index = self.env.open_db(b'block_index', txn=txn, dupsort=True, dupfixed=True)
      self.pubkey_index = self.env.open_db(b'pubkey_index', txn=txn, dupsort=True, dupfixed=True)
      self.pubkey_index_reversed = self.env.open_db(b'pubkey_index_reversed', txn=txn, dupsort=True, dupfixed=True)
      self.actions_list = self.env.open_db(b'actions_list', txn=txn)
      #if not txn.get(b'pool_size', db=self.pool):
      #  txn.put( b'pool_size', 0, db=self.pool) 



  def get_pool_prop(self, prop, r_txn):
    return r_txn.get( prop, db=self.pool) 

  def get_pool_size(self, r_txn):
    return self.get_pool_prop(b'size', r_txn=r_txn)

  def get_pool_current(self, r_txn):
    return self.get_pool_prop(b'current', r_txn=r_txn)

  def get_pool_top(self, r_txn):
    return self.get_pool_prop(b'top', r_txn=r_txn)

  def set_pool_prop(self, prop, value, w_txn):
    if not value==None:
        return w_txn.put( prop, value.to_bytes(4,'big'), db=self.pool) 
    else:
        return w_txn.delete( prop, db=self.pool) 

  def set_pool_size(self, value, w_txn):
    return self.set_pool_prop(b'size', value, w_txn=w_txn)

  def set_pool_current(self, value, w_txn):
    return self.set_pool_prop(b'current', value, w_txn=w_txn)

  def set_pool_top(self, value, w_txn):
    return self.set_pool_prop(b'top', value, w_txn=w_txn)

  def put(self, serialized_pubkey, serialized_privkey, w_txn):
    p1=w_txn.put( bytes(serialized_pubkey), bytes(serialized_privkey), db=self.main_db)    
  

  def put_output(self, output_index, output_params, w_txn):
    p1=w_txn.put( bytes(output_index), serialize_output_params(output_params), db=self.output)  
    self.add_block_new_output_association(output_params[0], output_index, w_txn=w_txn)  
    w_txn.put( _(self.last_action_num(r_txn=w_txn)), RECEIVED+bytes(output_index), db=self.actions_list)
    self.bump_action_num(w_txn=w_txn)

  def put_outgoing_output(self, output_index, output_params, w_txn):
    w_txn.put( bytes(output_index), serialize_output_params(output_params), db=self.generated_outputs)    

  def register_processed_output(self, output_index, w_txn):
    w_txn.put( _(self.last_action_num(r_txn=w_txn)), SENT+bytes(output_index), db=self.actions_list)
    self.bump_action_num(w_txn=w_txn)

  def get_output(self, output_index, r_txn):
    output_params = r_txn.get( bytes(output_index), db=self.output)    
    if not output_params:
       raise KeyError
    else:
      return deserialize_output_params( output_params)

  def get_outgoing_output(self, output_index, r_txn):
    output_params = r_txn.get( bytes(output_index), db=self.generated_outputs)    
    if not output_params:
       raise KeyError
    else:
      return deserialize_output_params( output_params)

  def pop_ser_output(self, output_index, w_txn):
    output_params = w_txn.pop( bytes(output_index), db=self.output)    
    if not output_params:
       raise KeyError
    else:
      return output_params

  def spend_output(self, output_index, block_height, w_txn):
    ser_output = self.pop_ser_output(output_index, w_txn=w_txn) # KeyError exception may be thrown here
    ser_spent_output = repack_ser_output_to_spent(ser_output, block_height)
    p1=w_txn.put( bytes(output_index), ser_spent_output, db=self.spent) 
    w_txn.put( _(self.last_action_num(r_txn=w_txn)), SPENT+bytes(output_index), db=self.actions_list)
    self.bump_action_num(w_txn=w_txn)

  def unspend_output(self, output_index, w_txn):
    ser_spent_output = w_txn.pop( bytes(output_index), w_txn=w_txn)
    if not ser_spent_output:
      raise KeyError
    ser_output = repack_ser_spent_output_to_unspent(ser_spent_output)
    p1=w_txn.put( bytes(output_index), ser_output, db=self.output) # TODO self.put_output?

  def get_spent_output(self, output_index, r_txn):
    output_params = r_txn.get( bytes(output_index), db=self.spent)    
    if not output_params:
       raise KeyError
    else:
      return deserialize_spent_output_params( output_params)


  def get_privkey_from_pool(self, w_txn):
      '''
        Privkey will be immideately removed from the pool.
      '''
      # pool is integer->pubkey key_value
      current = self.get_pool_current(r_txn=w_txn)
      if not current:
        raise Exception("Pool is empty")
      privkey = w_txn.get( w_txn.get(current, db=self.pool), db=self.main_db)
      current_int = int.from_bytes(current, 'big')
      top_int = int.from_bytes(self.get_pool_top(r_txn=w_txn), 'big')
      if current_int < top_int:
        self.set_pool_current(current_int+1, r_txn=w_txn)
      else:
        self.set_pool_current(None, w_txn=w_txn)
      self.set_pool_size(int.from_bytes(self.get_pool_size(r_txn=w_txn), 'big')-1, w_txn=w_txn)
      return privkey
        

  def add_privkey_to_pool(self, serialized_pubkey, serialized_privkey, w_txn):
    self.put(serialized_pubkey, serialized_privkey, w_txn=w_txn)
    pool_current = self.get_pool_current(r_txn=w_txn)
    if pool_current == None:
       self.set_pool_size(1, w_txn=w_txn)
       pool_current = b"\x00\x00\x00\x00"
       current = 0
       self.set_pool_current(current, w_txn=w_txn)
       self.set_pool_top(current, w_txn=w_txn)
    else:
       self.set_pool_size(int.from_bytes(self.get_pool_size(r_txn=w_txn),"big")+1, w_txn=w_txn)
       self.set_pool_top(int.from_bytes(self.get_pool_top(r_txn=w_txn),"big")+1, w_txn=w_txn)
    w_txn.put( self.get_pool_top(r_txn=w_txn), serialized_pubkey, db=self.pool)  


  def get_privkey(self, serialized_pubkey, r_txn):
    '''
      Note: Being get by pubkey, key pair is not checked wether it was in the pool or not.
    '''
    return r_txn.get(serialized_pubkey, db=self.main_db)
    #TODO if no such serialized_pubkey KeyError should be raised here

  def add_block_spent_output_association(self, block_height, output_index, w_txn):
    w_txn.put( _(block_height), b"\x01"+output_index, db=self.block_index, dupdata=True)
        

  def add_block_new_output_association(self, block_height, output_index, w_txn):
    w_txn.put( _(block_height), b"\x00"+output_index, db=self.block_index, dupdata=True)


  def remove_block_spent_output_association(self, block_height, output_index, w_txn):
    w_txn.delete( _(block_height), b"\x01"+output_index, db=self.block_index)


  def remove_block_new_output_association(self, block_height, output_index, w_txn):
    w_txn.delete( _(block_height), b"\x00"+output_index, db=self.block_index)


  def remove_all_outputs_created_in_block(self, block_height, w_txn):
    cursor = w_txn.cursor(db=self.block_index)
    if not cursor.set_key(_(block_height)):
      return #It's okay, block doesn't contain anything related to our wallet
    for assoc in cursor.iternext_dup():
      if assoc[:1] == b"\x00":
        self.pop_ser_output(assoc[1:], w_txn=w_txn)
        self.remove_block_new_output_association(block_height, assoc[1:], w_txn=w_txn)
      #NOTE: last n actions can also contain spent outputs, so remove_all_outputs_created_in_block call
      # without restore_all_outputs_spent_in_block call is unsafe
      #self.remove_last_actions(n=len(output_list), w_txn=w_txn)
          


  def restore_all_outputs_spent_in_block(self, block_height, w_txn):
      cursor = w_txn.cursor(db=self.block_index)
      output_list = []
      if not cursor.set_key(_(block_height)):
        return #It's okay, block doesn't contain anything related to our wallet
      for assoc in cursor.iternext_dup():
        if assoc[:1] == b"\x01":
          output_list.append(self.unspend_output(assoc[1:], w_txn=w_txn))
          self.remove_block_spent_output_association(block_height, assoc[1:], w_txn=w_txn)
      #NOTE: last n actions can also contain spent outputs, so remove_all_outputs_created_in_block call
      # without restore_all_outputs_spent_in_block call is unsafe
      #self.remove_last_actions(n=len(output_list), w_txn=w_txn)

  def search_output_everywhere(self, output_index, r_txn, guess=None):
      methods = [self.get_output, self.get_spent_output, self.get_outgoing_output]
      if not guess or guess==RECEIVED:
        pass
      elif guess==SPENT:
        methods = [self.get_spent_output, self.get_output, self.get_outgoing_output]
      elif guess==SENT:
        methods = methods[::-1]
      for method in methods:
        try:
          return method(output_index, r_txn)
        except KeyError:
          continue
      raise KeyError

  def remove_all_actions_in_last_block(self, block_height, w_txn):
    start = current = self.last_action_num(r_txn=w_txn)    
    num = 0
    while current>0:
      tp = w_txn.get(_(current-1), db = self.actions_list)
      act, output_index = tp[:1], tp[1:]
      try:
          height = self.search_output_everywhere(output_index, r_txn = w_txn)[0]
      except KeyError:
          pass
      if height<block_height:
        break
      current-=1
      num+=1
    self.remove_last_actions(num, w_txn)  
  
  def put_pubkey_output_assoc(self, pubkey, index, w_txn):
      w_txn.put( pubkey, index, db=self.pubkey_index, dupdata=True)
      w_txn.put( index, pubkey, db=self.pubkey_index_reversed, dupdata=True)

  def get_priv_and_pub_by_output_index(self, output_index, r_txn):
      pubkey = r_txn.get(output_index, db=self.pubkey_index_reversed)
      if not pubkey:
        raise KeyError
      priv = r_txn.get(pubkey, db=self.main_db)
      if not priv:
        raise KeyError
      return pubkey, priv

  def get_output_private_data(self, output_index, r_txn):
      pubkey = r_txn.get(output_index, db=self.pubkey_index_reversed)
      if not pubkey:
        raise KeyError
      priv = r_txn.get(pubkey, db=self.main_db)
      if not priv:
        raise KeyError
      output_params = r_txn.get( bytes(output_index), db=self.output)    
      if not output_params:
         raise KeyError
      params = deserialize_output_params( output_params)
      blinding_key = params[3]
      apc = params[4]
      return priv, blinding_key, apc 

  #TODO unused, remove?
  def add_sent_output(self, output, w_txn): 
    w_txn.put( _(self.last_action_num(r_txn=w_txn)), SENT+bytes(output_index), db=self.actions_list)
    self.bump_action_num(w_txn=w_txn)    

      
  def last_action_num(self, r_txn):
    lan = r_txn.get(b'last num', db = self.actions_list)
    if not lan:
      lan = 0
    else:
      lan = _d(lan)
    return lan

  def bump_action_num(self, w_txn, n=None):
    lan = self.last_action_num(r_txn=w_txn) if not n else n-1
    w_txn.put(b'last num', _(lan+1), db = self.actions_list)

  def remove_last_actions(self, n, w_txn):
    lan = self.last_action_num(r_txn=w_txn)
    for i in range(lan-1, lan-1-n, -1):
      w_txn.delete(_(i), db = self.actions_list)
    self.bump_action_num(w_txn=w_txn, n = lan-n)
        

  def generate_transaction_dict(self, n, r_txn):
      txdict = {} #Each element is dict {block_num: {'output':{params}}}
      current_action = self.last_action_num(r_txn=r_txn) - 1
      while current_action>=0 and len(txdict)<n:
        tp = r_txn.get(_(current_action), db = self.actions_list)
        action, output_index = tp[:1], tp[1:]
        soi = base64.b64encode(output_index).decode()
        current_action-=1
        output_params = None
        try:
          output_params = self.get_output(output_index, r_txn=r_txn)
        except KeyError:
          try:
            output_params = self.search_output_everywhere(output_index, r_txn = r_txn, guess=action)
          except KeyError:
            pass #TODO warning
            continue
        if action==SPENT:
          spend_height, created_height, lock_height, value, ser_blinding_key, serialized_apc, taddress = output_params
          if not spend_height in txdict:
            txdict[spend_height] = {}
          txdict[spend_height][soi] = {'lock_height':lock_height, 'value':value, 'address':taddress.decode(), 'type':'spent'}
        elif action==RECEIVED:
          if len(output_params)==6:
            created_height, lock_height, value, ser_blinding_key, serialized_apc, taddress = output_params
          elif len(output_params)==7: #It was received on created_height but later spent
            spend_height, created_height, lock_height, value, ser_blinding_key, serialized_apc, taddress = output_params
          if not created_height in txdict:
            txdict[created_height] = {}
          _type = 'received'
          if soi in txdict[created_height] and txdict[created_height][soi]['type']=='sent':
           _type = 'inner transfer'
          txdict[created_height][soi] = {'lock_height':lock_height, 'value':value, 'address':taddress.decode(), 'type':_type}
        elif action==SENT:
          created_height, lock_height, value, ser_blinding_key, serialized_apc, taddress = output_params
          if not created_height in txdict:
            txdict[created_height] = {}
          _type = 'sent'
          if soi in txdict[created_height] and txdict[created_height][soi]['type']=='received':
           _type = 'inner transfer'
          txdict[created_height][soi] = {'lock_height':lock_height, 'value':value, 'address':taddress.decode(), 'type':_type}
      return txdict  
        
            



from leer.core.storage.headers_storage import HeadersStorage

from leer.core.utils import encode_target, decode_target

from leer.core.parameters.constants import *
from math import exp, ceil

def next_target(_hash, headers_storage, rtx):
  span = 20
  if not headers_storage.has(_hash, rtx=rtx):
    raise
  header = headers_storage.get(_hash, rtx=rtx)
  if header.height<=span:
    return decode_target(*encode_target(initial_target))
  average_target =0
  runner = header
  for i in range(span):
    average_target +=runner.target/span
    runner = headers_storage.get(runner.prev, rtx=rtx)
  average_period = (header.timestamp - runner.timestamp)/span
  target = average_target * max(min(average_period/block_time, max_target_increase), max_target_decrease)
  if target > minimal_target:
    target = minimal_target
  target = decode_target(*encode_target(target))
  return target

def next_reward(_hash, headers_storage, rtx):
  span = 1024
  if _hash == b"\x00"*32:# 'prev' of genesis
    return max_reward(0)
  if not headers_storage.has(_hash, rtx=rtx):
    raise
  header = headers_storage.get(_hash, rtx=rtx)
  if header.height<=span:
    return max_reward(header.height+1)
  runner = header
  summ = 0
  for i in range(span):
    summ += runner.votedata.miner_subsidy_vote_int
    runner = headers_storage.get(runner.prev, rtx=rtx)
  return int(max_reward(header.height+1) * (summ/(255.*span)))

def max_reward(height):
  return int(initial_reward * exp(-float(height)/reward_decrease_halflife))



from leer.core.primitives.header import Header, PoPoW, VoteData
from leer.core.storage.txos_storage import TXOsStorage
from leer.core.chains.headers_manager import HeadersManager
from leer.core.storage.excesses_storage import ExcessesStorage
from leer.core.storage.headers_storage import HeadersStorage
from leer.core.primitives.transaction_skeleton import TransactionSkeleton
from leer.core.lubbadubdub.transaction import Transaction
from leer.core.lubbadubdub.ioput import IOput
from leer.core.lubbadubdub.offset_utils import sum_offset
from time import time
from leer.core.parameters.dynamic import next_reward, next_target
from leer.core.parameters.constants import initial_target
import functools

class Block():

  def __init__(self, storage_space, header=None, transaction_skeleton=None):
    self._header = header if header else Header()
    self.transaction_skeleton = transaction_skeleton if transaction_skeleton else TransactionSkeleton()
    self.tx=None
    self.storage_space = storage_space

  @property
  def header(self):
    try:
      return self._header
    except:
      self._header = Header()
      return self._header  

  @property
  def hash(self):
    return self.header.hash

  @property
  def partial_hash(self):
    return self.header.partial_hash

  def serialize(self, rtx, rich_block_format=False, max_size =40000):
    serialized=b""
    serialized += self.header.serialize()
    serialized += self.transaction_skeleton.serialize(rich_format=rich_block_format, max_size=max_size,
        full_tx = build_tx_from_skeleton(self.transaction_skeleton,\
                                         self.storage_space.txos_storage,\
                                         self.storage_space.excesses_storage,\
                                         self.header.height, self.header.version, rtx=rtx,\
                                         historical = True) if rich_block_format else None)
    return serialized

  @classmethod
  @functools.lru_cache(maxsize=40)
  def from_serialized(cls, serialized_block, storage_space):
    b = cls(storage_space=storage_space)
    b.deserialize(serialized_block)
    return b

  def deserialize(self, serialized):
    self.deserialize_raw(serialized)


  def deserialize_raw(self, serialized):
    serialized = self.header.deserialize_raw(serialized)
    serialized = self.transaction_skeleton.deserialize_raw(serialized, storage_space=self.storage_space)
    return serialized

  def non_context_verify(self, rtx):
    '''
      To verify block we need to
        0) check that header is known and valid
        1) verify transaction 
        2) check that transaction can be applied 
        3) #logic error, this is context validation# check that after tx applied to prev state, new state roots 
           and checksums coinside with block header
        4) check reward size (actually in can be checked on headers level)
    '''
    # stage 1
    assert self.storage_space.headers_storage.has(self.header.hash, rtx=rtx), "Block's header is unknown"
    #self.storage_space.headers_storage.context_validation(self.header.hash)
    assert not self.storage_space.headers_storage.get(self.header.hash, rtx=rtx).invalid, "Block's header is invalid. Reason: `%s`"%self.storage_space.headers_storage.get(self.header.hash, rtx=rtx).reason

    #currently during building we automatically check that tx can ba applied and tx is valid
    self.tx = build_tx_from_skeleton(self.transaction_skeleton, txos_storage=self.storage_space.txos_storage,
                                     excesses_storage=self.storage_space.excesses_storage,
                                     block_height=self.header.height, block_version = self.header.version, rtx=rtx, non_context = True)
    # stage 3 => should be moved to blockchain
    #commitment_root, txos_root = self.storage_space.txos_storage.apply_tx_get_merkles_and_rollback(tx)
    #excesses_root = self.storage_space.excesses_storage.apply_tx_get_merkles_and_rollback(tx)
    #assert [commitment_root, txos_root, excesses_root]==self.header.merkles

    # This is context validation too??? TODO
    assert self.tx.coinbase.value == (next_reward(self.header.prev, self.storage_space.headers_storage, rtx=rtx)+self.transaction_skeleton.relay_fee), "Wrong block subsidy"
    
    return True


  def __str__(self):
    return "Block< hash: %s..., height: %d, inputs: %d, outputs %d>"%(self.header.hash[:6], self.header.height
                    , len(self.transaction_skeleton.input_indexes),len(self.transaction_skeleton.output_indexes) )
    

def build_tx_from_skeleton(tx_skeleton, txos_storage, excesses_storage, block_height, block_version, rtx, historical=False, non_context = False):
  '''
    By given tx_skeleton and txos_storage return transaction.
    If transaction is invalid or any input/output isn't available exception will be raised.
    Optionally, if `historical` is True we will check output_indexes both in mempool and spent outputs. 
  '''
  tx=Transaction(txos_storage=txos_storage, excesses_storage=excesses_storage)
  for _i in tx_skeleton.input_indexes:
       tx.inputs.append(txos_storage.confirmed.get(_i, rtx=rtx))
  for _o in tx_skeleton.output_indexes:
       if historical or non_context:
         # About non_context: if we are on one branch and build block from another one
         # and this block contain output which is already commited on our branch (tx is
         # confirmed on both branches) we should get txo from confirmed storage
         try:
           tx.outputs.append(txos_storage.confirmed.find(_o, rtx=rtx))
         except:
           tx.outputs.append(txos_storage.mempool[_o])
       else:
         tx.outputs.append(txos_storage.mempool[_o])
  tx.additional_excesses = tx_skeleton.additional_excesses.copy()
  tx.updated_excesses = tx_skeleton.updated_excesses.copy()
  tx.mixer_offset = tx_skeleton.mixer_offset
  if historical or non_context:
    assert tx.non_context_verify(block_height=block_height)
  else:
    assert tx.verify(block_height=block_height, block_version = block_version, rtx=rtx)
  return tx

#To setup utils
def generate_genesis(tx, storage_space, wtx):
    '''
        1. spend inputs and add outputs and excesses from tx to storage
        2. calc new mercles
        3. generate header
        4. rollback outputs
    '''
    storage = storage_space.txos_storage
    excesses = storage_space.excesses_storage

    exc_merkle = excesses.apply_tx_get_merkles_and_rollback(tx, wtx=wtx) # it should be calced first, since we nned to calc address_excess_num_index
    merkles = storage.apply_tx_get_merkles_and_rollback(tx, wtx=wtx) + [exc_merkle]
    popow = PoPoW([])
    votedata = VoteData()
    target = initial_target
    full_offset = tx.mixer_offset
    header=Header(height = 0, supply=tx.coinbase.value, full_offset=full_offset, merkles=merkles, popow=popow, votedata=votedata, timestamp=int(time()), target=target, version=int(1), nonce=b"\x00"*16)
    
    tx_skeleton = TransactionSkeleton(tx=tx)
    new_block = Block(storage_space, header, tx_skeleton)
    return new_block




def generate_block_template(tx, storage_space, wtx, get_tx_from_mempool = True, timestamp = None):
    '''
        Generate block template: block is correct but nonce (by default) is equal to zero.
        Thus difficulty target (almost always) isn't met.
        arguments:
          tx [mandatory]: transaction which contains coinbase output. It also may contain other inputs and outputs.
          storage_space [mandatory] : -
          get_tx_from_mempool [optional, default True]: if get_tx_from_mempool, transaction from mempool will be merged to block_transaction. If this merge will produce invalid tx (for instance tx from mempool spends the same inputs as tx with coinbase), tx from mempool will be discarded.

        Inner logic:
        1. apply block_tx to txos_storage and excesses_storage
        2. calc new merkles
        3. generate header with new merkles
        4. generate block by appending tx_skeleton and new header
        5. rollback block_tx
    '''

    storage = storage_space.txos_storage
    excesses = storage_space.excesses_storage
    current_block = storage_space.blocks_storage.get(storage_space.blockchain.current_tip(rtx=wtx), rtx=wtx)
    if get_tx_from_mempool:
      try:
        tx = tx.merge(storage_space.mempool_tx.give_tx(), rtx=wtx)
      except:
        pass

    exc_merkle = excesses.apply_tx_get_merkles_and_rollback(tx, wtx=wtx) # it should be calced first, since we nned to calc address_excess_num_index
    merkles = storage.apply_tx_get_merkles_and_rollback(tx, wtx=wtx) + [exc_merkle]

    popow = current_block.header.next_popow()
    #We subtract relay fee, since coinbase value contain relay fees, but it isn't new money, but redistribution
    supply = current_block.header.supply + tx.coinbase.value - tx.calc_new_outputs_fee() - tx.relay_fee 
    height = current_block.header.height+1
    votedata = VoteData()
    target = next_target(current_block.hash, storage_space.headers_storage, rtx=wtx)    
    full_offset = sum_offset(current_block.header.full_offset,tx.mixer_offset)
    if not timestamp:
      timestamp = max(int(time()), storage_space.headers_storage.get(storage_space.blockchain.current_tip(rtx=wtx), rtx=wtx).timestamp+1)
    header=Header(height = height, supply=supply, full_offset=full_offset, merkles=merkles, popow=popow, votedata=votedata, timestamp=timestamp, target=target, version=int(1), nonce=b"\x00"*16)
    
    tx_skeleton = TransactionSkeleton(tx=tx)
    new_block = Block(storage_space, header, tx_skeleton)
    return new_block


class ContextBlock(Block):
  # TODO consider removing ContextBlock. For now we store all information about validity in ContextHeader
  # (it allows headers_manager to provide less useless paths).
  '''
    Wrapper of Block for inner storage. It contains contextual info about block: for instance is it valid in chain or not.
  '''
  def __init__(self, storage_space = None, block=None):
    if block:
      Block.__init__(self, storage_space= block.storage_space, header=block.header, transaction_skeleton=block.transaction_skeleton)
      if block.tx:
        self.tx=block.tx
    else:
      if not storage_space:
        raise TypeError("ContextBlock initialized without context")
      Block.__init__(self, storage_space)
    self.invalid = False
    self.reason = None

  def serialize_with_context(self):
    ser = super(ContextBlock, self).serialize(rtx=None) # We can pass None as rtx, since rtx is required for rich block serialization
    ser += int(self.invalid).to_bytes(1,'big')
    reason = self.reason if self.reason else ""
    ser += int(len(reason)).to_bytes(2,'big')
    ser += reason.encode('utf-8')
    return ser

  @classmethod
  @functools.lru_cache(maxsize=40)
  def from_serialized(cls, serialized_block, storage_space):
    b = cls(storage_space=storage_space)
    b.deserialize(serialized_block)
    return b

  def deserialize(self, serialized):
    self.deserialize_raw(serialized)

  def deserialize_raw(self, serialized):
    ser = super(ContextBlock, self).deserialize_raw(serialized)
    self.invalid, ser = bool(ser[0]), ser[1:]
    reason_len, ser = int.from_bytes(ser[:2], 'big'), ser[2:]
    self.reason, ser = ser[:reason_len].decode('utf-8'), ser[reason_len:]
    return ser
    
  def __str__(self):
    return "ContextBlock< hash: %s..., height: %d, inputs: %d, outputs %d, valid: %s, reason %s>"%(self.header.hash[:6], self.header.height
                    , len(self.transaction_skeleton.input_indexes),len(self.transaction_skeleton.output_indexes),
                    ("-" if self.invalid else '+'), self.reason  )





from time import time
from leer.core.primitives.block import Block, ContextBlock
from leer.core.primitives.header import Header
from leer.core.storage.blocks_storage import BlocksStorage, RollBack
from leer.core.storage.txos_storage import TXOsStorage
from leer.core.storage.headers_storage import HeadersStorage
from leer.core.storage.excesses_storage import ExcessesStorage
from leer.core.parameters.dynamic import next_reward
from leer.core.utils import DOSException

class Blockchain:
  def __init__(self, storage_space, notify_wallet=None):
    self.storage_space = storage_space
    self.awaited_blocks = {} # requested but not downloaded blocks
    self.storage_space.register_blockchain(self)
    self.download_queue = [] #Note: it's different from awaited blocks: download queue is queue of blocks still to be requested
    self.notify_wallet = notify_wallet

  def add_block(self, block, wtx, no_update=False):
    cb = ContextBlock(block=block)
    #TODO
    #if not block.hash in  self.awaited_blocks:
    #  raise Exception("Already have block")
    if self.storage_space.blocks_storage.has(block.hash, rtx=wtx):
      raise DOSException("Already have block")

    self.storage_space.blocks_storage.put(block.hash, cb, wtx=wtx)
    try:
      self.awaited_blocks.pop(block.hash)
    except:
      pass

    if not self.storage_space.blocks_storage.is_block_downloaded(block.hash, rtx=wtx, auto_download=True):
      pass
    else:
      try:
        assert block.non_context_verify(rtx=wtx)
      except Exception:
        self.storage_space.headers_manager.mark_subchain_invalid(block.hash, wtx=wtx, reason = "Block %s(h:%d) failed non-context validation"%(block.hash, block.header.height))
        return
      if not no_update: #During first syncing we want to get many blocks at once, so do not update after each
        self.update(wtx=wtx, reason="new block (verified)")
    
  def _add_block_to_chain(self, block_hash, wtx):
    '''
      To add block we need to apply tx both to txos storage and excesses.
      After that we need to set new state.
      Also we need to save info for rollback.
    '''
    block = self.storage_space.blocks_storage.get(block_hash, rtx=wtx)
    block.non_context_verify(rtx=wtx) #build tx from skeleton
    if not self.context_validation(block, wtx=wtx):
      ch = self.storage_space.headers_manager.mark_subchain_invalid(block.hash, wtx=wtx, reason = "Block %s(h:%d) failed context validation"%(block.hash, block.header.height))
      return self.update(wtx=wtx, reason="Detected corrupted block")    
    rb = RollBack()
    rb.prev_state = self.current_tip(rtx=wtx)
    # Note excesses_storage.apply_tx modidies transaction, in particular adds
    # context-dependent address_excess_num_index to outputs. Thus it should be applied before txos_storage.apply_tx
    excesses_num, rollback_updates = self.storage_space.excesses_storage.apply_tx(tx=block.tx, new_state=block_hash, wtx=wtx)  
    rollback_inputs, output_num = self.storage_space.txos_storage.apply_tx(tx=block.tx, new_state=block_hash, wtx=wtx)
    #Write to db
    burden_for_rollback = []
    for burden in block.tx.burdens:
      if not self.storage_space.txos_storage.confirmed.burden.get(burden[0], rtx=wtx):
        self.storage_space.txos_storage.confirmed.burden.put(burden[0], burden[1], wtx=wtx)
        burden_for_rollback.append((burden[0], burden[1]))
    # Rollback creation
    rb.pruned_inputs=rollback_inputs
    rb.updated_excesses = rollback_updates
    rb.num_of_added_outputs = output_num
    rb.num_of_added_excesses = excesses_num
    rb.burdens = burden_for_rollback
    self.storage_space.blocks_storage.put_rollback_object(block_hash, rb, wtx=wtx)
    self.storage_space.mempool_tx.update(rtx=wtx, reason="new block")
    if self.notify_wallet:
      self.notify_wallet("apply", block.tx, block.header.height)
    

  def _rollback(self, wtx):
    rb = self.storage_space.blocks_storage.pop_rollback_object(self.current_tip(rtx=wtx), wtx=wtx)
    h = self.current_height(rtx=wtx)
    self.storage_space.txos_storage.rollback(pruned_inputs=rb.pruned_inputs, num_of_added_outputs=rb.num_of_added_outputs, prev_state=rb.prev_state, wtx=wtx)
    self.storage_space.excesses_storage.rollback(num_of_added_excesses=rb.num_of_added_excesses, prev_state=rb.prev_state, rollback_updates=rb.updated_excesses, wtx=wtx)
    for burden in rb.burdens:
      self.storage_space.txos_storage.confirmed.burden.remove(burden[0], wtx=wtx)
    if self.notify_wallet:
      self.notify_wallet("rollback", rb, h)

  def clean_old_block_requests(self):
    to_delete = []
    for bh in self.awaited_blocks:
      if self.awaited_blocks[bh]+3600 < time():
         to_delete.append(bh)
    for bh in to_delete:
      self.awaited_blocks.pop(bh)

  def _lazy_ask_for_block(self, block_hash):
    self.download_queue.append(block_hash)

  def _download_queued_blocks(self):
    self.clean_old_block_requests()
    if len(self.awaited_blocks)>256:
      return
    to_download = []
    for bh in self.download_queue:
      if (not bh in to_download) and (not bh in self.awaited_blocks):
        self.awaited_blocks[bh]=time()
        to_download.append(bh)
    self.ask_for_blocks_hook(to_download)
    self.download_queue=[]

  def _ask_for_blocks(self, block_hash):
    self.clean_old_block_requests()
    if len(self.awaited_blocks)>256 or (block_hash in self.awaited_blocks):
      return
    self.awaited_blocks[block_hash]=time()
    self.ask_for_blocks_hook(block_hash) #It should be set by user
     

  def context_validation(self, block, wtx):
    assert block.header.prev == self.current_tip(rtx=wtx)
    block.tx.verify(block_height=self.current_height(rtx=wtx), block_version = block.header.version, rtx=wtx, skip_non_context=True)
    excesses_root = self.storage_space.excesses_storage.apply_tx_get_merkles_and_rollback(block.tx, wtx=wtx)
    commitment_root, txos_root = self.storage_space.txos_storage.apply_tx_get_merkles_and_rollback(block.tx, wtx=wtx)
    if not [commitment_root, txos_root, excesses_root]==block.header.merkles:
      return False
    '''excesses = block.tx.additional_excesses + list(block.tx.updated_excesses.values())
    excesses_indexes = [e.index for e in excesses]
    for i in block.tx.inputs:
      if self.storage_space.txos_storage.confirmed.burden.has(i.serialized_index, rtx=wtx):
        required_commitment = txos.storage.confirmed.burden.get(i.serialized_index, rtx=wtx)
        required_index =
        if (not required_index in excesses_indexes) and (not self.storage_space.excesses_storage.has_index(required_index)):
          return False
    for excess in excesses:
      if  self.current_height(rtx=wtx)>0:
        prev_block_props = {'height': self.current_height(rtx=wtx), 
                         'timestamp': self.storage_space.headers_storage.get(self.current_tip(rtx=wtx), rtx=wtx).timestamp}
      else:
        prev_block_props = {'height':0, 'timestamp':0}
      burden_list = []
      excess_lookup_partial = partial(excess_lookup, rtx=wtx, tx=block.tx, excesses_storage = self.storage_space.excesses_storage)
      output_lookup_partial = partial(output_lookup, rtx=wtx, tx=block.tx, txos_storage = self.storage_space.txos_storage)
      result = execute(script = excess.message,
                       prev_block_props = prev_block_props,
                       excess_lookup = excess_lookup_partial,
                       output_lookup = output_lookup_partial,
                       burden = burden_list)
      if not result:
        return False
    '''
    if block.header.height>0:
      subsidy = next_reward(block.header.prev, self.storage_space.headers_storage, rtx=wtx)
      if not self.storage_space.headers_storage.get(block.header.prev, rtx=wtx).supply + \
           subsidy - \
           block.transaction_skeleton.calc_new_outputs_fee(is_block_transaction=True) == block.header.supply:
        return False
      if not block.tx.coinbase.value == subsidy + block.tx.relay_fee: 
        # Note we already check coinbase in non_context_check, but using tx_skeleton info
        # Since information in tx_skeleton may be forged, this check is not futile
        return False
      
    return True

  def current_tip(self, rtx):
    ts,es = self.storage_space.txos_storage.confirmed.get_state(rtx=rtx), self.storage_space.excesses_storage.get_state(rtx=rtx)
    es=None if es == b"" else es #TODO
    assert ts==es
    if ts==None:
      return b"\x00"*32
    return ts

  def current_height(self,rtx):
    current_tip = self.current_tip(rtx=rtx)
    if current_tip==b"\x00"*32:
      current_height = -1
    else:
      current_height = self.storage_space.blocks_storage.get(current_tip,rtx=rtx).header.height
    return current_height

  def update(self, wtx, reason=None):
    current_tip = self.current_tip(rtx=wtx)
    max_steps = 2048
    actions = self.storage_space.headers_manager.next_actions(current_tip, rtx=wtx, n = max_steps)
    # before doing anything we should check that we have enough info (downloaded blocks)
    # to move to a better state (state with higher height)
    current_height = self.current_height(rtx=wtx)
    good_path=None
    for path in actions:
      if good_path:
        break
      for ind, step in enumerate(path):
        if step[0]=="ADDBLOCK":
          if not  self.storage_space.blocks_storage.has(step[1], rtx=wtx):
            #Try to download as much blocks as possible and break
            for _step in path:
              if _step[0]=="ADDBLOCK" and (not self.storage_space.blocks_storage.has(_step[1], rtx=wtx)):
                self._lazy_ask_for_block(_step[1])
            self._download_queued_blocks()
            break
          if (not self.storage_space.blocks_storage.is_block_downloaded(step[1], rtx=wtx)):
            #lets look forward and ask more txos to be downloaded
            for action, block_hash in path[ind+1:]:
               if self.storage_space.blocks_storage.has(block_hash, rtx=wtx):
                 self.storage_space.blocks_storage.is_block_downloaded(block_hash, rtx=wtx)
            break
          else:
            if self.storage_space.blocks_storage.get(step[1], rtx=wtx).header.height>current_height or \
               self.storage_space.blocks_storage.get(step[1], rtx=wtx).header.height-current_height>=max_steps:
              good_path=path
              break
    if not good_path:
      return
    progress = self.process_path(good_path, wtx=wtx)
    if progress:# and (not good_path==actions[0]):
      # workaround for situations when branch with best known header is not available:
      # in this case next_actions will return only one step for alternative branch.
      # details in HeadersManager.next_actions
      self.update(wtx=wtx, reason="recursive check")


  def process_path(self, path, wtx):
    progress = False
    for step in path:
      action, block_hash = step
      if action=="ROLLBACK":
        while self.current_tip(rtx=wtx)!=block_hash:
          self._rollback(wtx=wtx)
          #TODO Add some checks here to prevent rolling back to genesis in case of any mistakes
      if action=="ADDBLOCK":
        if (not self.storage_space.blocks_storage.has(block_hash, rtx=wtx)) or \
           (not self.storage_space.blocks_storage.is_block_downloaded(block_hash, rtx=wtx)):
          # We decide that path is good if we have enough downloaded blocks to get to current_height+1. Still, the path
          # may contain more steps and for some of those steps we may not have downloaded blocks. So stop now.
          break
        try:
          self._add_block_to_chain(step[1], wtx=wtx)
          progress = True #at least one block was added
        except AssertionError:
          break
    return progress

  def _forget_top(self, wtx):
    """
      This function is used for tests only: currently we can generate blocks only on top of known ones.
      Thus, to collide two forks (one of which is previously unknown to blockchain) we need first to generate forks
      and then to force blockchain to forget it.
      This function deletes data not only from blockchain but also from blocks storage and txos_storage.
      Note, after forgeting top block, previous block became blockchain tip. If there is a known longer branch, 
      blockchain will not be switched to it.
    """
    ct =self.current_tip(rtx=wtx)
    block = self.storage_space.blocks_storage.get(ct, rtx=wtx)
    self._rollback(wtx=wtx)
    self.storage_space.blocks_storage.forget_block(ct, wtx=wtx)
    for _o in block.transaction_skeleton.output_indexes:
      self.storage_space.txos_storage.mempool.remove_by_index(_o)
      


  def is_block_in_main_chain(self, block_hash, rtx):
    header = self.storage_space.headers_storage.get(block_hash,rtx=rtx)
    return header_hash == self.storage_space.heders_manager.find_ancestor_with_height(self.current_tip, header.height, rtx=rtx) 
    




VERSION = "0.2.1"
CODENAME = "Buccinator"
OVERALL_DEVELOPMENT_STATUS = "Beta"

P2P_TRANSPORT_STATUS = "Ready"
LUBBADUBDUB_STATUS = "Ready"
ASSETS_STATUS = "Not Implemented"
BLOCKCHAIN_STATUS = "Ready"
WALLET_STATUS = "Beta"
RPC_STATUS = "Beta"



import json, re
from os.path import *
from os import urandom

def commentjson_loads(text, **kwargs):
    '''
      (c) -> https://github.com/vaidik/commentjson/
      We do not have commentjson as dependency to decrease exposure
    '''
    regex = r'\s*(#|\/{2}).*$'
    regex_inline = r'(:?(?:\s)*([A-Za-z\d\.{}]*)|((?<=\").*\"),?)(?:\s)*(((#|(\/{2})).*)|)$'
    lines = text.split('\n')
    for index, line in enumerate(lines):
        if re.search(regex, line):
            if re.search(r'^' + regex, line, re.IGNORECASE):
                lines[index] = ""
            elif re.search(regex_inline, line):
                lines[index] = re.sub(regex_inline, r'\1', line)
    minified = '\n'.join(lines)
    try:
      return json.loads(minified, **kwargs)    
    except Exception as e:
     err = str(e)
     have_line = err.find("line")
     if have_line>0:
       n = int(err[have_line+5:].split(" ")[0])
       print("Error in line `%s`"%lines[n])
       raise Exception("Problem with config")
     else:
       raise e


def get_public_IP():
    data = str(urlopen('http://checkip.dyndns.com/').read())
    pat="Current IP Address: "
    data = data[data.find(pat)+len(pat):]
    data = data.split("<")[0]
    return data

def get_static_key(filename):
  comment =\
  '''
    P2p protocol which leer uses requires private key for encrypting communication.
    This key should be known prior communication. Nevertheless, this key is independent
    of any wallet or blockchain cryptography.
  '''
  try:
    with open(filename, "r") as f:
      for l in f.readlines():
        if not len(l) or l[0]=="#":
          continue
        key = int(l)
  except IOError:
    raise
    #if not path.isdir(filename):
    #  makedirs(join(expanduser("~"), ".leertest"))
    key = int.from_bytes(urandom(32), "big")
    with open(filename, "w") as f:
      comment = "\n".join(["#"+i for i in comment.split("\n")])
      f.write(comment+'\n')
      f.write("%d"%key)
  return key


def main(config):
  import asyncio
  import logging
  import sys

  from urllib.request import urlopen
  from functools import partial

  from secp256k1_zkp import PrivateKey, PublicKey
  from leer.syncer import Syncer
  from leer.transport.network_manager import NM_launcher
  from leer.rpc.rpc_manager import RPCM_launcher
  from leer.core.core_loop import core_loop
  from leer.notification.notification_center import notification_center_launcher
  from leer.wallet.wallet import wallet as wallet_launcher
  import multiprocessing
  import time

  from os import path, makedirs
  logging.basicConfig(level=logging.ERROR, format='%(asctime)s %(name)s %(levelname)s:%(message)s')
  
  if "advertised_host" in config["p2p"] and config["p2p"]["advertised_host"] == "autodetect":
     config["p2p"]["advertised_host"] = get_public_IP()
  config["p2p"]["lspriv"]=get_static_key(expanduser(config["p2p"]["lspriv_file"]))
  if "location" in config:
    for t in config["location"]:
      config["location"][t] = expanduser(config["location"][t])

  async def start_server(config, delay_before_connect=5):
    syncer=Syncer()
    nm = multiprocessing.Process(target=NM_launcher, args=(syncer, config))
    nm.start()
    rpcm = multiprocessing.Process(target=RPCM_launcher, args=(syncer, config))
    rpcm.start()
    core = multiprocessing.Process(target=core_loop, args=(syncer, config))
    core.start()
    await asyncio.sleep(delay_before_connect)
    notifications = multiprocessing.Process(target=notification_center_launcher, args=(syncer, config))
    notifications.start()
    if "wallet" in config and config["wallet"]:
      wallet = multiprocessing.Process(target=wallet_launcher, args=(syncer, config))
      wallet.start()
    for node in config['bootstrap_nodes']:
      #print("Require connection from %d to %d: %s:%s:%s"%(server_id, node, 'localhost', p2p_port_by_id(node), pub_key_by_id(node)))
      syncer.queues['NetworkManager'].put(
        {'action':'open connection', 
         'host':node['host'],
         'port':node['port'],
         'static_key':node['pub'], 
         'id':int((time.time()*1e5)%1e5),
         'sender': "RPC"})
    while True:
      await asyncio.sleep(1)
  
  loop = asyncio.get_event_loop()
  asyncio.ensure_future(start_server(config))
  loop.run_forever()




if __name__ == '__main__':
    import sys
    relativ_config_path = sys.argv[1]
    config = None
    with open(relativ_config_path, "r") as f:
      raw_config = f.read()
      config = commentjson_loads(raw_config)
    
    # execute only if run as the entry point into the program
    main(config)






from leer.core.storage.txos_storage import TXOsStorage
from leer.core.primitives.block import Block, ContextBlock
import shutil, os, time, lmdb, math
import functools

class BlocksStorage:
  __shared_states = {}
  def __init__(self, storage_space, wtx):
    path = storage_space.path
    if not path in self.__shared_states:
      self.__shared_states[path]={}
    self.__dict__ = self.__shared_states[path]

    self.storage = BlocksDiscStorage(path, env=storage_space.env, wtx=wtx)
    self.storage_space = storage_space
    self.storage_space.register_blocks_storage(self)
    self.download_queue = []
    
  @functools.lru_cache(maxsize=15)
  def get(self, _hash, rtx):
    serialized_context_block = self.storage.get_by_hash(_hash, rtx=rtx)
    if not serialized_context_block:
      raise KeyError(_hash)
    cblock = ContextBlock.from_serialized(serialized_context_block, self.storage_space)
    return cblock

  def put(self, _hash, block, wtx):
    self.storage.put(_hash, block.serialize_with_context(), wtx=wtx)

  def has(self, _hash, rtx):
    return self.storage.has(_hash, rtx=rtx)  

  def is_block_downloaded(self, _hash, rtx, auto_download=True):
    asked = False
    result = True
    block = self.get(_hash, rtx=rtx)
    for output in block.transaction_skeleton.output_indexes:
      if not self.storage_space.txos_storage.known(output, rtx=rtx):
        result = False
        if auto_download:
          self._ask_for_txout(output)
          asked = True
    if asked:
      self._ask_for_txouts()#flushing requests
    return result

  def get_rollback_object(self, _hash, rtx):
    serialized_rollback = self.storage.get_rollback_object(_hash, rtx=rtx)
    rb=RollBack()
    rb.deserialize_raw(serialized_rollback)
    return rb

  def pop_rollback_object(self, _hash, wtx):
    serialized_rollback = self.storage.pop_rollback_object(_hash, wtx=wtx)
    rb=RollBack()
    rb.deserialize_raw(serialized_rollback)
    return rb

  def put_rollback_object(self, _hash, rollback, wtx):
    self.storage.put_rollback_object(_hash, rollback.serialize(), wtx=wtx)



  def _ask_for_txout(self, txout):
    if not txout in self.download_queue:
      self.download_queue.append(txout)

  def _ask_for_txouts(self):
    self.ask_for_txouts_hook(self.download_queue)
    self.download_queue = []

  def forget_block(self, _hash, wtx):
    self.storage.delete_block_by_hash(_hash, wtx=wtx)



class RollBack:
  def __init__(self):
    self.pruned_inputs = [] #revert objects
    self.num_of_added_outputs = None
    self.num_of_added_excesses = None
    self.prev_state = None
    self.updated_excesses = []
    self.burdens = []

  def serialize_bytes(self, _bytes):
    return len(_bytes).to_bytes(2,"big")+_bytes

  def deserialize_bytes_raw(self, serialized):
    _slen, serialized = serialized[:2], serialized[2:]
    _len = int.from_bytes(_slen, "big")
    _bytes, serialized = serialized[:_len], serialized[_len:]
    return _bytes, serialized

  def serialize(self):
    # each pruned input is (TXOS[num, _index, obj], Commitment[num, _index, obj])
    # num is serialized 5bytes integer, other members have arbitrary size
    serialized_pruned_inputs=b""
    serialized_pruned_inputs+=len(self.pruned_inputs).to_bytes(2,"big")
    for _t, _c in self.pruned_inputs:
      for _i in [_t,_c]:
        num, _index, obj = _i
        serialized_pruned_inputs += num.to_bytes(5,"big") + self.serialize_bytes(_index) + self.serialize_bytes(obj)

    serialized_excess_updates=b""
    serialized_excess_updates+=len(self.updated_excesses).to_bytes(2,"big")
    for update in self.updated_excesses:
      num, _index, obj = update
      serialized_excess_updates += num + self.serialize_bytes(_index) + self.serialize_bytes(obj)

    serialized_nums = self.num_of_added_outputs.to_bytes(2,"big") + self.num_of_added_excesses.to_bytes(2,"big")
    serialized_burdens_len = len(self.burdens).to_bytes(2,"big")
    serialized_burdens = serialized_burdens_len + b"".join([i[0] for i in self.burdens])
    version=b"\x01"
    serialized_state_id = self.serialize_bytes(self.prev_state)
    summary_len = len(serialized_pruned_inputs)+len(serialized_nums) + len(serialized_burdens)+len(version)+len(serialized_state_id) + len(serialized_excess_updates)
    serialized = summary_len.to_bytes(4,"big") + \
                 version + \
                 serialized_pruned_inputs + \
                 serialized_nums + \
                 serialized_excess_updates + \
                 serialized_burdens + \
                 serialized_state_id
    return serialized

  def deserialize_raw(self, serialized):
    _slen, serialized = serialized[:4], serialized[4:]
    _len = int.from_bytes(_slen, "big")
    data, residue = serialized[:_len], serialized[_len:]
    version, data = data[:1], data[1:]
    if not version==b"\x01":
      raise
    _s_pruned_inputs_num, data = data[:2], data[2:]
    pruned_inputs_num = int.from_bytes(_s_pruned_inputs_num, "big")
    for i in range(pruned_inputs_num):
      _snum, data = data[:5], data[5:]
      _num = int.from_bytes(_snum, "big")
      _index, data = self.deserialize_bytes_raw(data)
      _obj, data = self.deserialize_bytes_raw(data)
      _txout = [_num,_index,_obj]
      _snum, data = data[:5], data[5:]
      _num = int.from_bytes(_snum, "big")
      _index, data = self.deserialize_bytes_raw(data)
      _obj, data = self.deserialize_bytes_raw(data)
      _comm = [_num,_index,_obj]
      self.pruned_inputs.append([_txout, _comm])
    _no, _ne, data = data[:2], data[2:4], data[4:]
    self.num_of_added_outputs = int.from_bytes(_no,"big")
    self.num_of_added_excesses = int.from_bytes(_ne, "big")
    _s_updates_num, data = data[:2], data[2:]
    updates_num = int.from_bytes(_s_updates_num, "big")
    for i in range(updates_num):
      _snum, data = data[:5], data[5:]
      _index, data = self.deserialize_bytes_raw(data)
      _obj, data = self.deserialize_bytes_raw(data)
      self.updated_excesses.append((_snum, _index, _obj))
    _s_burdens_num, data = data[:2], data[2:]
    burdens_num = int.from_bytes(_s_burdens_num, "big")
    for i in range(burdens_num):
      next_burden, data= data[:65], data[65:]
      self.burdens.append((next_burden, None))
    self.prev_state, data = self.deserialize_bytes_raw(data)
    return residue


class BlocksDiscStorage:
  def __init__(self, dir_path, env, wtx):
    self.dir_path = dir_path

    self.env = env
    self.main_db = self.env.open_db(b'blocks_main_db', txn=wtx, dupsort=False) # block_hash -> serialized_contextblock
    self.revert_db = self.env.open_db(b'blocks_revert_db', txn=wtx, dupsort=False) # block_hash -> object_for_reverting

  def put(self, _hash, serialized_block, wtx):
    p1=wtx.put( bytes(_hash), bytes(serialized_block), db=self.main_db, dupdata=False, overwrite=True)

  def update(self, _hash, serialized_block, wtx):
    wtx.put(bytes(_hash), bytes(serialized_block), db=self.main_db, dupdata=False, overwrite=True)

  def get_by_hash(self, _hash, rtx):
    return rtx.get(bytes(_hash), db=self.main_db)

  def get_rollback_object(self, _hash, rtx):
    return rtx.get(bytes(_hash), db=self.revert_db)

  def pop_rollback_object(self, _hash, wtx):
    return wtx.pop(bytes(_hash), db=self.revert_db)

  def put_rollback_object(self, _hash, serialized_rollback_object, wtx):
    return wtx.put(bytes(_hash), bytes(serialized_rollback_object), db=self.revert_db)

  def has(self, _hash, rtx):
    return bool(self.get_by_hash(_hash, rtx=rtx))

  def delete_block_by_hash(self, _hash, wtx):
    wtx.delete( bytes(_hash), db=self.main_db)
    
  




import asyncio
import logging
import time


from secp256k1_zkp import PrivateKey, PublicKey

from .node import Node
from lnoise import Key
logger = logging.getLogger(__name__)
from .messages import message_id, inv_message_id
from uuid import uuid4


class NetworkManager:
  def __init__(self, loop, syncer, config):
    self.loop = loop
    self.config = config 
    self.syncer = syncer
    self.global_message_queue = syncer.queues['NetworkManager']
    self.load_from_disc()
    #for node in self.nodes.values():
    #  self.connect_to(node)
    self.loop.call_soon(self.check_global_message_queue)
    self.nodes={}
    self.reconnect_list = {}
    self.connecting = [] # nodes that are connecting, but not connected yet
    self.server = asyncio.start_server(self.handle_connection, config['p2p']['host'], config['p2p']['port'], loop=loop)
    self.tasks = []
    self.up = True
    asyncio.ensure_future(self.server, loop=loop)
    asyncio.ensure_future(self.reconnect_loop(), loop=loop)

    #set logging
    default_log_level = logging.INFO;
    if "logging" in config:#debug, info, warning, error, critical
      loglevels = { "debug":logging.DEBUG, "info":logging.INFO, "warning":logging.WARNING, "error":logging.ERROR, "critical":logging.CRITICAL}
      if "base" in config["logging"] and config["logging"]["base"] in loglevels:
        logger.setLevel(loglevels[config["logging"]["base"]])
      if "network" in config["logging"] and config["logging"]["network"] in loglevels:
        #its ok to rewrite
        logger.setLevel(loglevels[config["logging"]["network"]])

  def load_from_disc(self):
    lspriv=self.config['p2p']['lspriv']
    s=Key(key=PrivateKey(lspriv.to_bytes(32,'big'), raw=True))
    our_node_params = { 'network': 
                           {
                             'host': self.config['p2p']['host'], 
                             'port':self.config['p2p']['port'],
                             'advertised_host': self.config['p2p'].get('advertised_host', self.config['p2p']['host']), 
                             'advertised_port': self.config['p2p'].get('advertised_port', self.config['p2p']['port']), 
                           }, 
                        'static_full_key':s}
    self.our_node = Node(None, our_node_params, self.loop, None)
    logger.info("Our node public key %s"%s.pubkey())
    self.nodes={}

  def save_to_disc(self):
    pass #TODO

  async def connect_to(self, host_port_tuple, static_key=None):
    if not host_port_tuple in self.nodes:
      if not static_key:
        raise Exception("Cannot connect to new node without static key")
      host, port = host_port_tuple
      new_node_params = { 'network': {'host':host, 'port':port}, 'static_key':static_key}
      node = Node(self.our_node, new_node_params, self.loop, self.handle_message)
      while host_port_tuple in self.connecting:
        await asyncio.sleep(1)
        if host_port_tuple in self.nodes:
          return
      self.connecting.append(host_port_tuple)
      await node.connect()
      self.connecting.remove(host_port_tuple)

  async def handle_connection(self, reader, writer):
    extra_info = writer.get_extra_info('peername')
    host, port = extra_info[:2]
    logger.info("New connection from %s %s"%(str(host), str(port)))
    params = {'network':{'host':host, 'port':port}}
    new_node = Node(self.our_node, params, self.loop, self.handle_message)
    await new_node.accept_connection(reader, writer)

  async def handle_message(self, node, _type, message):

    if _type == "close": # TODO explanation that message can be thrown by Node-object itself in on_disconnect
      if message[0] == 0:
        if (node.host, node.port) in self.nodes:
          self.nodes.pop((node.host, node.port))
        #accidental disconnet, set to reconnect
        node_params = (node.advertised_host, node.advertised_port)
        if not node_params in self.reconnect_list:
          if node.advertised_static_key: #we can't reconnect without static key
            self.reconnect_list[node_params] = {'static_key':node.advertised_static_key, 'last_try_time':time.time(), 'try':0}
        else:
          self.reconnect_list[node_params]['last_try_time']=time.time()
          self.reconnect_list[node_params]['try']+=1 
        if node_params in self.nodes:
          self.nodes.pop(node_params)

    if _type == "init":
      node.deserialize_params(message, remote=True)
      self.nodes[(node.host, node.port)]=node 
      if (node.advertised_host, node.advertised_port) in self.reconnect_list:
        self.reconnect_list.pop((node.advertised_host, node.advertised_port))
        #XXX possible attack here: if an attacker wants to exclude a node (segment)
        # from the network. (s)he can DDoS this node and then reconnect it to all other
        # nodes. If (s)he will advertise the attacked
        # node host and port as its own, disconnected nodes will not try to
        # reconnect. At the same time. if the attacker simulates multiple connections
        # to the attacked node from previously connected nodes (again, advertising fake
        # host and port), attacked node will not try to reconnect to disconnected
        # nodes either. It is a difficult attack which requires enormous DDoS
        # abilities, knowing of network topology, and absence of fresh coming nodes.
        # Nevertheless, this issue should be revisited

    if _type == "give nodes":
      node_list_to_send = []
      for k,v in self.nodes.items():
        if not v==node: 
          node_list_to_send.append(v.serialize_params())
      message = b"\x00\04"+len(node_list_to_send).to_bytes(2, "big") #TODO move byte operation unde node interface
      for node_to_send in node_list_to_send:
        message += len(node_to_send).to_bytes(2, "big") + node_to_send
      await node.send(message) 

    if _type == "take nodes":
      try:
        node_list_len, r = message[:2], message[2:]
        node_list_len = int.from_bytes(node_list_len, "big")
        nodes=[]

        local_in_connection=[]
        try:
          # Node appears in known_nodes only after `init` message
          # So if the information about the node will reach us multiple times before
          # the first successfull connection, we will try to connect to it multiple times
          # self.in_connection is used to prohibit such behavior
          self.in_connection
        except:
          self.in_connection = []

        for i in range(node_list_len):
          _node_len, r = r[:2], r[2:]
          _node_len = int.from_bytes(_node_len, "big")
          _node, r =  r[:_node_len], r[_node_len:]
          new_node = Node(self.our_node, None, self.loop, self.handle_message, serialized_params=_node)
          #nodes.append(new_node.connect())
          if not (new_node.host, new_node.port) in self.get_known_nodes():
            if not (new_node.host, new_node.port) in self.in_connection: 
              if not (new_node.host, new_node.port) == (self.our_node.host, self.our_node.port):
                if not new_node.static_key.serialize() == self.our_node.static_key.serialize(): # mirror replay
                  #await new_node.connect()
                  task = new_node.connect()
                  nodes.append(task)
                  self.in_connection.append((new_node.host, new_node.port))
                  local_in_connection.append((new_node.host, new_node.port))
        if len(nodes):     
          await asyncio.wait(nodes)
          for _node in local_in_connection:
            self.in_connection.remove(_node)
      except Exception as e:
        print("exception nm", e)
        pass #TODO DoS protection

    if _type == "give next headers":
      try:
        _ser_num, from_hash = message[:2], message[2:32+2]
        num = int.from_bytes(_ser_num,"big")
        self.syncer.queues["Blockchain"].put({'action': "give next headers",
                                              'id':str(uuid4()), "num": num, 
                                              "from": bytes(from_hash), "node":(node.host, node.port), 'sender':"NetworkManager"})
      except Exception as e:
        print("exception nm", e)
        pass #TODO DoS protection

    if _type == "take the headers":
      try:
        _ser_num, headers = message[:2], message[2:]
        num = int.from_bytes(_ser_num,"big")
        self.syncer.queues["Blockchain"].put({'action': 'take the headers',
                                              'id':str(uuid4()), "num": num, 
                                              "headers": bytes(headers), "node":(node.host, node.port), 'sender':"NetworkManager"})
      except Exception as e:
        print("exception nm", e)
        pass #TODO DoS protection

    if _type == "give blocks":
      try:
        _ser_num, block_hashes = message[:2], message[2:]
        num = int.from_bytes(_ser_num,"big")
        self.syncer.queues["Blockchain"].put({'action': 'give blocks',
                                              'id':str(uuid4()), "num": num, 
                                              "block_hashes": bytes(block_hashes), "node":(node.host, node.port), 'sender':"NetworkManager"})
      except Exception as e:
        print("exception nm", e)
        pass #TODO DoS protection

    if _type == "take the blocks":
      try:
        _ser_num, blocks = message[:2], message[2:]
        num = int.from_bytes(_ser_num,"big")
        self.syncer.queues["Blockchain"].put({'action': 'take the blocks',
                                              'id':str(uuid4()), "num": num, 
                                              "blocks": bytes(blocks), "node":(node.host, node.port), 'sender':"NetworkManager"})
      except Exception as e:
        print("exception nm", e)
        pass #TODO DoS protection

    if _type == "give the txos":
      try:
        _ser_num, txos_hashes = message[:2], message[2:]
        num = int.from_bytes(_ser_num,"big")
        self.syncer.queues["Blockchain"].put({'action': 'give txos',
                                              'id':str(uuid4()), "num": num, 
                                              "txos_hashes": bytes(txos_hashes), "node":(node.host, node.port), 'sender':"NetworkManager"})
      except Exception as e:
        print("exception nm", e)
        pass #TODO DoS protection

    if _type == "take the txos":
      try:
        _ser_num, message = message[:2], message[2:]
        num = int.from_bytes(_ser_num,"big")
        txos_hashes, txos_lengths, txos = message[:num*65], message[num*65:num*65+num*2], message[num*65+num*2:] 
        self.syncer.queues["Blockchain"].put({'action': 'take the txos',
                                              'id':str(uuid4()), "num": num, 
                                              "txos_hashes": bytes(txos_hashes), "txos_lengths": bytes(txos_lengths),
                                              "txos": bytes(txos), "node":(node.host, node.port), 'sender':"NetworkManager"})
      except Exception as e:
        print("exception nm", e)
        pass #TODO DoS protection


    if _type == "take tip info":
      try:
        _ser_height, message = message[:4], message[4:]
        tip, prev_hash, message = message[:32], message[32:64], message[64:]
        _ser_td = message[:32]
        height, total_difficulty = int.from_bytes(_ser_height, "big"), int.from_bytes(_ser_td, "big"), 
        self.syncer.queues["Blockchain"].put({'action': 'take tip info',
                                              'id':str(uuid4()), "height": height, "prev_hash": bytes(prev_hash),
                                              "tip":bytes(tip), "total_difficulty":total_difficulty, "node":(node.host, node.port), 'sender':"NetworkManager"})
      except Exception as e:
        print("exception nm", e)
        pass #TODO DoS protection

    if _type == "find common root":
      try:
        self.syncer.queues["Blockchain"].put({'action': 'find common root',
                                              'id':str(uuid4()), "serialized_header": bytes(message),
                                              "node":(node.host, node.port), 'sender':"NetworkManager"})
      except Exception as e:
        print("exception nm", e)
        pass #TODO DoS protection

    if _type == "find common root response":
      try:
        header_hash, serialized_len, message = message[:32], message[32:33], message[33:]
        _len = int.from_bytes(serialized_len, "big") # SHouldn't we just use message[32] ?
        known_headers = message[:_len]
        self.syncer.queues["Blockchain"].put({'action': 'find common root response',
                                              'id':str(uuid4()), "header_hash": bytes(header_hash),
                                              'flags_num':_len, "known_headers": bytes(known_headers),
                                              "node":(node.host, node.port), 'sender':"NetworkManager"})
      except Exception as e:
        print("exception nm", e)
        pass #TODO DoS protection

    if _type == "take TBM transaction":
      try:
        serialized_mode, serialized_skeleton = message[:2], message[2:]
        self.syncer.queues["Blockchain"].put({'action': 'take TBM transaction',
                                              'id':str(uuid4()), "tx_skel": bytes(serialized_skeleton),
                                              "mode": int.from_bytes(serialized_mode, "big"),
                                              "node":(node.host, node.port), 'sender':"NetworkManager"})
      except Exception as e:
        print("exception nm", e)
        pass #TODO DoS protection

    if _type == "give TBM transaction":
      try:
        self.syncer.queues["Blockchain"].put({'action': 'give TBM transaction',
                                              "node":(node.host, node.port), 'sender':"NetworkManager"})
      except Exception as e:
        print("exception nm", e)
        pass #TODO DoS protection





  def get_known_nodes(self):
    nodes = [(node.advertised_host,node.advertised_port) for k,node in self.nodes.items()]
    nodes = nodes+list(self.nodes)
    return set(nodes)
    

  def check_global_message_queue(self):
      while not self.global_message_queue.empty():
          message = self.global_message_queue.get()
          action = message['action']
          if action == "open connection":
              host, port, static_key = message['host'], message['port'], PublicKey(message['static_key'], raw=True)
              coro = self.connect_to( (host, port), static_key=static_key)
              asyncio.ensure_future(coro, loop=self.loop)
          if action == "get connections num":
              _id = message['id']
              request_source = message['sender'] 
              self.syncer.queues[request_source].put({'id':_id, 'result':len(self.nodes)})

          if action == "give intrinsic nodes list": #Not list anymore, considr renaming TODO
              _id = message['id']
              request_source = message['sender'] 
              nodes_info = list(self.nodes.keys())
              self.syncer.queues[request_source].put({'id':_id, 'result':nodes_info})

          if action == "give nodes list": #Not list anymore, considr renaming TODO
              _id = message['id']
              request_source = message['sender'] 
              nodes_info = {(self.nodes[node].advertised_host, self.nodes[node].advertised_port):self.nodes[node].static_key.serialize() for node in self.nodes}
              self.syncer.queues[request_source].put({'id':_id, 'result':nodes_info})

          if action == "give my node":
              _id = message['id']
              request_source = message['sender'] 
              nodes_info = {(self.our_node.advertised_host, self.our_node.advertised_port): self.our_node.static_key.serialize()}
              self.syncer.queues[request_source].put({'id':_id, 'result':nodes_info})

          if action == "take the headers":
            num, headers, node_params = message["num"], message["headers"], message["node"]
            if not node_params in self.nodes:
              continue
            message_to_send = inv_message_id["take the headers"]
            message_to_send += num.to_bytes(2,"big")
            message_to_send += headers
            coro = self.nodes[node_params].send(message_to_send)
            asyncio.ensure_future(coro, loop=self.loop)
            self.syncer.queues[message['sender']].put({'id':message['id'], 'result':'processed'})
          if action == "take the blocks":
            num, blocks, node_params = message["num"], message["blocks"], message["node"]
            if not node_params in self.nodes:
              continue
            message_to_send = inv_message_id["take the blocks"]
            message_to_send += num.to_bytes(2,"big")
            message_to_send += blocks
            coro = self.nodes[node_params].send(message_to_send)
            asyncio.ensure_future(coro, loop=self.loop)
            self.syncer.queues[message['sender']].put({'id':message['id'], 'result':'processed'})
          if action == "take the txos":
            num, txos, txos_hashes, txos_lengths, node_params = message["num"], message["txos"], message["txos_hashes"], message["txos_lengths"], message["node"]
            if not node_params in self.nodes:
              continue
            message_to_send = inv_message_id["take the txos"]
            message_to_send += num.to_bytes(2,"big")
            message_to_send += txos_hashes
            message_to_send += txos_lengths
            message_to_send += txos
            coro = self.nodes[node_params].send(message_to_send)
            asyncio.ensure_future(coro, loop=self.loop)
            self.syncer.queues[message['sender']].put({'id':message['id'], 'result':'processed'})
          if action == "give blocks":
            num, blocks_hashes, node_params = message["num"], message["block_hashes"], message["node"]
            if not node_params in self.nodes:
              continue   
            message_to_send = inv_message_id["give blocks"]
            message_to_send += num.to_bytes(2,"big")
            message_to_send += blocks_hashes
            coro = self.nodes[node_params].send(message_to_send)
            asyncio.ensure_future(coro, loop=self.loop)
            self.syncer.queues[message['sender']].put({'id':message['id'], 'result':'processed'})
          if action == "give next headers":
            num, from_hash, node_params = message["num"], message["from"], message["node"]
            if not node_params in self.nodes:
              continue   
            message_to_send = inv_message_id["give next headers"]
            message_to_send += num.to_bytes(2,"big")
            message_to_send += from_hash
            coro = self.nodes[node_params].send(message_to_send)
            asyncio.ensure_future(coro, loop=self.loop)
            self.syncer.queues[message['sender']].put({'id':message['id'], 'result':'processed'})
          if action == "give txos":
            num, txos_hashes, node_params = message["num"], message["txos_hashes"], message["node"]
            if not node_params in self.nodes:
              continue   
            message_to_send = inv_message_id["give the txos"]
            message_to_send += num.to_bytes(2,"big")
            message_to_send += txos_hashes
            coro = self.nodes[node_params].send(message_to_send)
            asyncio.ensure_future(coro, loop=self.loop)            
            self.syncer.queues[message['sender']].put({'id':message['id'], 'result':'processed'})
          if action == "take tip info":
            logger.info("Take tip info")
            height, tip, prev_hash, total_difficulty, node_params = message["height"], message["tip"], message["prev_hash"], message["total_difficulty"], message["node"]
            if not node_params in self.nodes:
              continue
            message_to_send = inv_message_id["take tip info"]
            message_to_send += height.to_bytes(4,"big")
            message_to_send += tip
            message_to_send += prev_hash
            message_to_send += total_difficulty.to_bytes(32,"big")
            coro = self.nodes[node_params].send(message_to_send)
            asyncio.ensure_future(coro, loop=self.loop)
            self.syncer.queues[message['sender']].put({'id':message['id'], 'result':'processed'})
          if action == "find common root":
            serialized_header, node_params = message["serialized_header"], message["node"]
            if not node_params in self.nodes:
              continue
            message_to_send = inv_message_id["find common root"]
            message_to_send += serialized_header
            coro = self.nodes[node_params].send(message_to_send)
            asyncio.ensure_future(coro, loop=self.loop)            
            self.syncer.queues[message['sender']].put({'id':message['id'], 'result':'processed'})
          if action == "find common root response":
            header_hash, known_headers, _len, node_params = message["header_hash"], message["known_headers"], message['flags_num'], message["node"]
            if not node_params in self.nodes:
              continue
            message_to_send = inv_message_id["find common root response"]
            message_to_send += header_hash
            message_to_send += _len.to_bytes(1,"big")
            message_to_send += known_headers
            coro = self.nodes[node_params].send(message_to_send)
            asyncio.ensure_future(coro, loop=self.loop)            
            self.syncer.queues[message['sender']].put({'id':message['id'], 'result':'processed'})

          if action == "take TBM transaction":
            mode, serialized_tx_skel, node_params = message["mode"], message["tx_skel"], message["node"]
            if not node_params in self.nodes:
              continue
            message_to_send = inv_message_id["take TBM transaction"]
            message_to_send += mode.to_bytes(2,"big")
            message_to_send += serialized_tx_skel
            coro = self.nodes[node_params].send(message_to_send)
            asyncio.ensure_future(coro, loop=self.loop)            
            self.syncer.queues[message['sender']].put({'id':message['id'], 'result':'processed'})

          if action == "give TBM transaction":
            node_params = message["node"]
            if not node_params in self.nodes:
              continue
            message_to_send = inv_message_id["give TBM transaction"]
            coro = self.nodes[node_params].send(message_to_send)
            asyncio.ensure_future(coro, loop=self.loop)            
            self.syncer.queues[message['sender']].put({'id':message['id'], 'result':'processed'})

          '''if action == "send ping":
              for node in self.nodes:
                coro = node.send( "ping 0")
                asyncio.ensure_future(coro, loop=self.loop)'''
          if action == "stop":
            self.loop.stop()
            return
      if self.up:
        self.loop.call_later(0.5, self.check_global_message_queue)
    
  async def reconnect_loop(self):
    def try_num_to_delay(try_num):
      if try_num==0:
        return 0
      delay = 30*(2**try_num)
      if delay > 1*3600:
        delay = 1*3600
      return delay
    while self.up:
      for node_params in self.reconnect_list:
        ltt = self.reconnect_list[node_params]['last_try_time']
        t = self.reconnect_list[node_params]['try']
        if time.time()>ltt+try_num_to_delay(t):
          asyncio.ensure_future(self.connect_to( node_params, static_key=self.reconnect_list[node_params]['static_key']))
      await asyncio.sleep(5)



def NM_launcher(syncer, config):
  loop = asyncio.new_event_loop()
  asyncio.set_event_loop(loop)
  NM = NetworkManager(loop, syncer, config)
  loop.run_forever()
  




from leer.core.lubbadubdub.constants import default_generator, default_generator_ser, generators
from leer.core.lubbadubdub.ioput import IOput
from leer.core.lubbadubdub.address import Excess
from leer.core.parameters.constants import output_creation_fee


class TransactionSkeleton:
  def __init__(self, tx=None):
    self.input_indexes = [] 
    self.output_indexes = []
    self.output_relay_fees = []
    self.additional_excesses = []
    self.updated_excesses = {}
    self.mixer_offset = 0
    self.tx = tx
    self.version = 0
    if tx:
      for _i in tx.inputs:
        self.input_indexes.append(_i.serialized_index)
      for _o in tx.outputs:
        self.output_indexes.append(_o.serialized_index)
        self.output_relay_fees.append(_o.relay_fee)
      self.additional_excesses = tx.additional_excesses.copy()
      self.updated_excesses = tx.updated_excesses.copy()
      self.mixer_offset = tx.mixer_offset

  def serialize(self, rich_format=False, max_size =40000, full_tx = None):
    #TODO we messed up a lot here with storage-space-free skeletons
    # Basically idea was that tx_skeleton is independent from storage_space, however in this case
    # we cannot serialize in rich format: outputs are stored only in storage_space. Moreover tx_skel
    # with all outputs actually contains the same info as tx. Now tx_skel needs space only if we want 
    # to serialize in rich_format, but since we cannot import build_tx_from_skeleton (cyclic import)
    # we pass full_tx into tx_skel.serialize . Probably the whole concept of 
    # transaction/transaction_skeleton should be reconsidered. 
    full_tx = full_tx if full_tx else self.tx
    if rich_format and not full_tx:
      raise Exception("Full_tx is required for serialization in rich format")
    serialization_array = []
    version_byte = ((self.version<<1)+int(rich_format)).to_bytes(1,"big") #lowest bit sets rich/not_rich format. Other bits are used for version
    serialization_array.append(version_byte)
    serialization_array.append(len(self.input_indexes).to_bytes(2, "big"))
    serialization_array.append(len(self.output_indexes).to_bytes(2, "big"))
    serialization_array.append(len(self.additional_excesses).to_bytes(2, "big"))
    serialization_array+=(self.input_indexes)
    serialization_array+=(self.output_indexes)
    serialization_array+=[i.to_bytes(4, "big") for i in self.output_relay_fees]
    serialization_array+=([ e.serialize() for e in self.additional_excesses])
    serialization_array+=([ self.updated_excesses[i].serialize() for i in self.input_indexes])
    serialization_array+=[self.mixer_offset.to_bytes(32,"big")]
    tx_skel_size = sum([len(i) for i in serialization_array])
    if rich_format and tx_skel_size<max_size:
      #we start with coinbase, because receiver definetely doesn't have this data
      txouts_data = b""
      txouts_count = 0
      tx = full_tx
      if tx.coinbase:
        serialized_coinbase = tx.coinbase.serialize()
        if len(serialized_coinbase)+tx_skel_size<max_size-2:
          txouts_data += serialized_coinbase
          txouts_count +=1
      for _o in tx.outputs:
          if _o.is_coinbase:
            continue
          serialized_output = _o.serialize()
          if tx_skel_size+len(txouts_data)+len(serialized_output)<max_size-2:
            txouts_data += serialized_output
            txouts_count +=1
          else:
            break
      if not txouts_count:
        #we don't have enough space even for one output
        serialization_array[0] = (self.version<<1).to_bytes(1,"big")
      else:
        serialization_array.append(txouts_count.to_bytes(2,"big"))
        serialization_array.append(txouts_data)

    return b"".join(serialization_array)

  def deserialize(self, serialized):
    self.deserialize_raw(serialized)
      

  def deserialize_raw(self, serialized, storage_space=None):
    if len(serialized)<1:
      raise Exception("Not enough bytes for tx skeleton version marker")
    serialized, ser_version = serialized[1:], serialized[0]
    rich_format = ser_version & 1
    self.version = ser_version >> 1
    if not self.version in [0]:
      raise Exception("Unknown tx_sceleton version")
    if len(serialized)<2:
      raise Exception("Not enough bytes for tx skeleton inputs len")
    serialized, _len_i = serialized[2:], int.from_bytes(serialized[:2], "big")
    if len(serialized)<2:
      raise Exception("Not enough bytes for tx skeleton outputs len")
    serialized, _len_o = serialized[2:], int.from_bytes(serialized[:2], "big")

    if len(serialized)<2:
      raise Exception("Not enough bytes for tx skeleton additional excesses len")
    serialized, _len_ae = serialized[2:], int.from_bytes(serialized[:2], "big")

    serialized_index_len = IOput().index_len
    for i in range(_len_i):
      if len(serialized)<serialized_index_len:
        raise Exception("Not enough bytes for tx skeleton' input index %d len"%i)
      _input_index, serialized  = serialized[:serialized_index_len], serialized[serialized_index_len:]
      self.input_indexes.append(_input_index)
    for i in range(_len_o):
      if len(serialized)<serialized_index_len:
        raise Exception("Not enough bytes for tx skeleton' output index %d len"%i)
      _output_index, serialized  = serialized[:serialized_index_len], serialized[serialized_index_len:]
      self.output_indexes.append(_output_index)

    for i in range(_len_o):
        if len(serialized)<4:
          raise Exception("Not enough bytes for tx skeleton' output relay fee %d len"%i)
        ser_relay_fee, serialized  = serialized[:4], serialized[4:]
        self.output_relay_fees.append(int.from_bytes(ser_relay_fee, "big"))

    for i in range(_len_ae):
      e = Excess()
      serialized  = e.deserialize_raw(serialized)
      self.additional_excesses.append(e)
    for i in range(_len_i):
      e = Excess()
      serialized  = e.deserialize_raw(serialized)
      self.updated_excesses[self.input_indexes[i]]=e
    if len(serialized)<32:
      raise Exception("Not enough bytes for mixer offset")
    self.mixer_offset, serialized = int.from_bytes(serialized[:32], "big"), serialized[32:]

    if not self.verify():
      #TODO consider renmaing verify to validate_excesses or make exception text more general
      raise Exception("Additional excesses are not signed properly")

    if rich_format and storage_space:
      txouts_num_serialized, serialized = serialized[:2], serialized[2:]
      txouts_num = int.from_bytes(txouts_num_serialized, "big")
      for _ in range(txouts_num):
        output = IOput()
        serialized = output.deserialize_raw(serialized)
        if not (output.serialized_index in self.output_indexes):
          raise Exception("Unknown output in rich txskel data") 
        storage_space.txos_storage.mempool[output.serialized_index]=output

    return serialized

  def verify(self):
    #We cannot verify sum to zero by tx_scel, since tx_scel doesn't contain address_excesses
    return True

  def calc_new_outputs_fee(self, is_block_transaction):
    return ( len(self.output_indexes) - len(self.input_indexes) - int(bool(is_block_transaction)) )*output_creation_fee

  @property
  def relay_fee(self):
    return sum(self.output_relay_fees)

  def __eq__(self, another_one):
    return self.serialize() == another_one.serialize()




class KeyValueStorage:
  def __init__(self, name, env, wtx, dupsort=False):
    self.name = name.encode("utf-8")
    self.env = env
    self.main_db = self.env.open_db(self.name + b'_main_db', txn=wtx, dupsort=dupsort)

  def put(self, key, value, wtx, dupdata=False):
      return wtx.put( bytes(key), bytes(value), db=self.main_db, dupdata=dupdata)

  def get(self, key, rtx):
    return rtx.get( bytes(key), db=self.main_db)

  def remove(self, key, wtx, value=None):
    return wtx.delete( bytes(key), value= value, db=self.main_db)

  def has(self, key, rtx):
    return not (self.get( bytes(key), rtx=rtx)==None)

